#ifndef _RDBJOIN_H_
#define _RDBJOIN_H_

#include "RdbRelation.h"


// A join condition is one (equality) condition  between two relations
// (possiblity the same relation.
// It represents the subscripts into a join.
// should probably add a condition operator...


class RdbJoinCondition
{
  public:

    RdbJoinCondition( int r1, int c1, int r2, int c2 );

    // RdbJoinCondition(const RdbJoinCondition&); default ok
    // RdbJoinCondition& operator =(const RdbJoinCondition&); default ok

    // access methods:

    int Rel1() const;
    int Col1() const;
    int Rel2() const;
    int Col2() const;

  private:

    // rel1.col1 == rel2.col2
    int rel1;			
    int col1;
    int rel2;
    int col2;

};

/***************************************************************************/

typedef PtrList<RdbRelation> RdbRelList;
typedef List<RdbJoinCondition> RdbJoinCondList;


class RdbJoin
{
  public:

    RdbJoin(const RdbRelList& newRelList, const RdbJoinCondList& newJoinCond);
				// create a join on these relations, with
				// this set of join conditions.
				// All relations must be in the same database
				// (this is not checked, however...).

    RdbJoin(const RdbJoin& other); // copy

    RdbSystem& System() const;	// database this join happens in

    RdbDatabase& Database() const; // database this join happens in

    RdbQuery* NewQuery(const BitVector& adorn);
				// create a new query on this join

    String SystemName() const;	// system name this join is in

    String DatabaseName() const; // database name this join is in

    String Schema() const;	// schema that results from this join
				// rel(col,col,col),rel(col,col)...

    int RelCount() const;	// number of relation in the join

    int JoinCount() const;	// how many join conditions?

    int Arity() const;		// max attributes in the result

    RdbRelation& Rel(int i) const; // gimme the i'th relation 

    String RelName(int i) const; // relation i's name

    void ColNumberToRelCol(int colNum, int& rel, int& col) const;
				// convert colNum to (rel,col)

    String ColName(int i) const; // column i's name

    String QualifiedColName(int rel,int col) const; // relname.col

    String QualifiedColName(int col) const; // relname.col

    String VarQualifiedColName(int rel,int col) const; // tuplevar.col

    String VarQualifiedColName(int col) const; // tuplevar.col

    String VarJoinString(int j) const; // tv1.col1 = tv2.col2

    RdbJoin& operator =(const RdbJoin& other); // copy


  private:

    RdbRelList relList;	// list of relations in the join

    RdbJoinCondList joinCond; // list of join conditions in join


};


#include "RdbJoin.i"

#endif // _RDBJOIN_H_
