#ifndef _RDBQUERYVERSIONS_H_
#define _RDBQUERYVERSIONS_H_

// RdbQueryVersions maintains a set of queries on one join of relations.
// Each query has a different adornment set.
// (Selection on different attributes, or leaves some attributes off the end)

#include "List.h"
#include "RdbQuery.h"
#include "RdbJoin.h"
#include "BitVector.h"
#include "String.h"


typedef DList<RdbQuery> RdbQueryList;

class RdbQueryVersions
{
  public:

    RdbQueryVersions(const RdbJoin& newJoin);
				// Establish versions on this set relations,
				// attributes, and join conditions

    virtual ~RdbQueryVersions();

    RdbQuery& Lookup(BitVector& adorn);
				// find the query with this adornment, 
				// create if not found

    // not impl: RdbQuery& Delete(BitVector& adorn);

    RdbSystem& System() const; // the system the queries are in

    RdbDatabase& Database() const; // the database the queries are in

    String SystemName() const;	// Name of system all queries connect to

    String DatabaseName() const; // Name of database all queries connect to

    String Schema() const;	// rel(col,,col),rel(col,,col)...

    int Arity() const;		// number of attributes in join (max)

    const RdbJoin& Join() const; // access to join

    const RdbQueryList& QueryList() const; // access to queryList

  private:

    RdbQueryList queryList;	// all the queries currently created

    RdbJoin join;		// the set to query

    RdbQueryVersions(const RdbQueryVersions&); // not allowed

    RdbQueryVersions& operator =(const RdbQueryVersions&); // not allowed

};

#include "RdbQueryVersions.i"


#endif // _RDBQUERYVERSIONS_H_
