
inline int PtrListBase::Size() const
{
    return count;
}


inline void PtrListBase::DeleteAll()
{
    count = 0;
}


inline int PtrListBase::ValidSubscript(int i) const
{
    return( i >= 0 && i < count );
}


inline void* PtrListBase::operator []( int i ) const
{
    assert( ValidSubscript(i) );
    return list[i];
}


inline void* PtrListBase::Set(int i, void* element)
{
    assert( ValidSubscript(i) );
    void* old = list[i];
    list[i] = element;
    return old;
}


//---------------------------------------------------------------------------

template <class T>
inline PtrList<T>::PtrList(int startSize = 10)
: PtrListBase(startSize)
{
}


template <class T>
inline PtrList<T>::PtrList( const PtrList<T>& other )
: PtrListBase(other)
{
}


template <class T>
inline PtrList<T>::~PtrList()
{
}


template <class T>
inline PtrList<T>& PtrList<T>::operator =(const PtrList<T>& other)
{
    PtrListBase::operator =(other);
    return *this;
}


template <class T>
inline T* PtrList<T>::operator []( int i ) const
{
    return (T*)PtrListBase::operator [](i);
}


template <class T>
inline int PtrList<T>::Append( T* element, int n )
{
    return PtrListBase::Append(element, n);
}


template <class T>
inline int PtrList<T>::Insert(T* element, int loc)
{
    return PtrListBase::Insert(element, loc);
}


template <class T>
inline T* PtrList<T>::Set(int i, T* element)
{
    return (T*)(PtrListBase::Set(i, element));
}


template <class T>
inline T* PtrList<T>::Delete(int i)
{
    return (T*)(PtrListBase::Delete(i));
}


//---------------------------------------------------------------------------

template <class T>
inline DList<T>::DList(int startSize = 10)
: PtrList<T>(startSize)
{
}


template <class T>
inline DList<T>::DList(const DList<T>& other)
: PtrList<T>(other)
{ 
}


template <class T>
inline DList<T>::~DList()
{
    DeleteAll(); 
}


template <class T>
inline DList<T>& DList<T>::operator =(const DList<T>& other)
{
    DeleteAll();
    PtrList<T>::operator =(other);
    return *this;
}


template <class T>
inline void DList<T>::Delete(int i)
{
    delete PtrList<T>::Delete(i);
}


template <class T>
inline void DList<T>::DeleteAll()
{
    for( int i = Size() -1 ; i >= 0 ; i-- ) {
	Delete(i);
    }
}


template <class T>
inline void DList<T>::Set(int i, T* element)
{
    delete PtrList<T>::Set(i, element);
}


//---------------------------------------------------------------------------

template <class T>
inline List<T>::List(int startSize = 10)
: DList<T>(startSize)
{
}


template <class T>
inline List<T>::List(const List<T>& other)
: DList<T>(other.Size())
{
    *this = other;
}


template <class T>
inline List<T>::~List()
{
    DeleteAll();
}


template <class T>
inline List<T>& List<T>::operator =(const List<T>& other)
{
    DeleteAll();
    for( int i = 0 ; i < other.Size() ; i++ ) {
	Append( other[i] );
    }
    return *this;
}


template <class T>
inline int List<T>::Append(const T& element)
{
    return DList<T>::Append( new T(element) );
}


template <class T>
inline int List<T>::Insert(const T& element, int loc)
{
    return DList<T>::Insert( new T(element), loc );
}


template <class T>
inline void List<T>::Set(int i, const T& element)
{
    DList<T>::Set(i, new T(element));
}


template <class T>
inline T& List<T>::operator []( int i ) const
{
    return *((T*)DList<T>::operator [](i));
}


//---------------------------------------------------------------------------

template <class T>
inline SortedList<T>::SortedList(int startSize = 10)
: List<T>(startSize)
{
}


template <class T>
inline SortedList<T>::SortedList(const SortedList<T>& other)
: List<T>(other)
{
}


template <class T>
inline SortedList<T>::~SortedList()
{
}


template <class T>
inline SortedList<T>& SortedList<T>::operator =(const SortedList<T>& other)
{
    this->List<T>::operator=(other);
    return *this;
}


template <class T>
inline int SortedList<T>::Insert(const T& element)
{
    int s = Size();
    for( int i = 0 ; i < s ; i++ ) {
	if( element <= (*this)[i] ) {
	    break;
	}
    }	  
    return List<T>::Insert(element, i);
}


//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

