// This may look like C code, but it is really -*- C++ -*-

//
// Coral.S - implementation of tcl "coral" command
// 
//**************************************************************************
// File Name:  Coral.S
//             Switched to tcl by Shaun Flisakowski using
//             treesh as a template, from Allan Brighton.
// Date: 
// File:       Coral.S
// Comments: 
//**************************************************************************

extern  "C" {
#include <tcl.h>
} // extern "C"

#include <iostream.h>
#include <fstream.h>

#include <stdlib.h>
#include <string.h>

#include "Coral.h"
#include "Coral_Utils.h"
#include <new.h>
#include <assert.h>
#include <ctype.h>
#include <sys/param.h>
#include <unistd.h>
#include <assert.h>
#include "Interface.h"

void bomb(void);

extern int errno;

#include "CoralInit.h"
#include "Globals.h"
#include "ParseStack.h"
#include "CoralError.h"
#include "Util.h"
#include "BuiltinUtil.h"
#include "Database.h"
#include "Interface.h"
#include "ConnectArray.h"

Coral *coral = NULL;
Tcl_HashTable table;
Tcl_Interp *glob_interp;

char *create_entry(char *prefix, void *store);

// *************************************************************************
void parse_all(){
    
    while(!Cor_parseStack->is_empty()){

        Cor_processInputStack();        
    }
}

// *************************************************************************
// Inputs: program name
// Outputs: Constructor
// Comments: Attempt to load the coralFuncs file from the 
//             CORALROOT/DIRECTORY directory.
// *************************************************************************

Coral::Coral(char *pname)
{
    if (!pname)
        pname = "coralsh";

    init_coral(pname);

    Cor_exEnv->C_print_prompt_default = 0;    // Turn off the prompt

    set_new_handler(bomb);
    //parse_all();
}

// *************************************************************************

_coral_export int module_name(char *);

int module_name(char *mname)
{
    Tcl_VarEval(glob_interp, "set_module ", mname, NULL );
    return(0);
}

// *************************************************************************
//    explain_break:
//            This function is a Coral "relation", it supplies a break-out
//         point in the middle of consulting the explain data so that
//            the GUI can process events (at a breakpoint)
// *************************************************************************

_coral_export int explain_break(int);

int explain_break(int i)
{
    char snum[8];

    sprintf(snum,"%d",i);

    Tcl_VarEval(glob_interp, "run_explain ", snum, NULL );

        /* The "Continue" command returns here to consult the rest
                of the file */

    return(0);
}

// *************************************************************************
// Inputs: Argc, Argv - A list of files to consult 
// Returns: An error or nothing (on success). 
// Added By: Shaun Flisakowski Sept 7 1994
// Comments: The return value (arg_id) is passed in interp->result
//                 consultFile is found in misc/Util.C
// *************************************************************************

int Coral::Consult(Tcl_Interp *interp, int argc, char **argv)

{
    int exist, err = 0;
    ifstream *fs;
    Tcl_DString tmp;
    char *filename;
    int   prev_exec_mode = Cor_exEnv->C_exec_mode;

    Tcl_DStringInit(&tmp);

    while(argc-- > 0){

        if ((filename = Tcl_TildeSubst(interp,*argv++,&tmp)) == NULL)
            return(TCL_ERROR);

        if (strcmp(filename,"-") == 0){

            Cor_exEnv->C_exec_mode = COR_DOING_INTERPRET;

            if (! Cor_consultFile(cin_strm, "stdin")){
                Tcl_AppendResult(interp,"Consult of stdin (pipe): failed.", 0);
                err = 1;
                continue;
            }

            Cor_parseStack->parse(cerr_strm,0);
            Cor_exEnv->C_exec_mode = (ExecutionMode) prev_exec_mode;

        } else {

            \[
                consult(($string) $filename).
            \]
        }
    }

    parse_all();

    Tcl_DStringFree(&tmp);

    if (err)
        return(TCL_ERROR);
    else
        return(TCL_OK);
}

// *************************************************************************
// Inputs: Argc, Argv - A list of strings to parse
// Returns: An error or nothing (on success). 
// Added By: Shaun Flisakowski Sept 7 1994
// Comments: The return value (arg_id) is passed in interp->result
//                 consultBuff is found in misc/Util.C
// *************************************************************************

int Coral::Parse(Tcl_Interp *interp, int argc, char **argv)

{
    int err = 0;

    while(argc-- > 0){

        if (! Cor_consultBuff(*argv++, "buffer")){
            Tcl_AppendResult(interp,"Consult of buffer failed.", 0);
            err = 1;
            continue;
        }
    }

    Cor_parseStack->parse(cerr_strm,0);

    if (err)
        return(TCL_ERROR);
    else
        return(TCL_OK);
}

// *************************************************************************
// Inputs: A tuple_id and an index.
// Returns: An arg_id for the index argument of tuple (if any). 
// Added By: Shaun Flisakowski Sept 6 1994
// Comments: The return value (arg_id) is passed in interp->result
// *************************************************************************

int Coral::GetArg(Tcl_Interp *interp, char *tuple, char *index)

{
    int indx;
    Tcl_HashEntry *entry;

    if ((entry = Tcl_FindHashEntry(&table,tuple)) == NULL) {
        Tcl_AppendResult(interp,tuple,": does not exist. ",0);
        return(TCL_ERROR);
    }

    if (Tcl_GetInt(interp,index,&indx) != TCL_OK)
        return(TCL_ERROR);

    if (strncmp(tuple,"tuple",5) == 0) {

        Cor_Tuple *tup = (Cor_Tuple *) Tcl_GetHashValue(entry);
        if (indx >= tup->arity()){
            Tcl_AppendResult(interp,"Not that many args in tuple.",0);
            return(TCL_ERROR);
        }

        Cor_CArg *arg = new Cor_CArg((*tup)[indx]);
        Tcl_AppendResult(interp,create_entry("arg",(void *) arg),0);

    } else {

        Tcl_AppendResult(interp,tuple,": can only get arg of a tuple.",0);
        return(TCL_ERROR);
    }

    return(TCL_OK);
}

// *************************************************************************
// Inputs: An open scan_id.
// Returns: None.
// Added By: Shaun Flisakowski Sept 6 1994
// Comments: Closes this scan.
// *************************************************************************

int Coral::CloseScan(Tcl_Interp *interp, char *scan_id)
{
    Tcl_HashEntry *entry;

    if ((entry = Tcl_FindHashEntry(&table,scan_id)) == NULL) {
        Tcl_AppendResult(interp,scan_id,": does not exist.",0);
        return(TCL_ERROR);
    }

    if (strncmp(scan_id,"scan",4) == 0) {

        Cor_ScanDesc *scan = (Cor_ScanDesc *) Tcl_GetHashValue(entry);
        delete scan;
        Tcl_DeleteHashEntry(entry);

    } else {

        Tcl_AppendResult(interp,scan_id,": can only close an open scan.",0);
        return(TCL_ERROR);
    }

    return(TCL_OK);
}

// *************************************************************************
// Inputs: argc, argv for a list of arg_id's and tuple_ids to delete. 
// Returns: None.
// Added By: Shaun Flisakowski Sept 6 1994
// Comments: Deletes the args and tuples from the hash table.
// *************************************************************************

int Coral::DeleteItems(Tcl_Interp *interp, int argc, char **argv)
{
    Tcl_HashEntry *entry;
    int err = 0;

    while(argc-- > 0) {

        char *item = *argv++;

        if ((item == NULL) || (*item == '\0'))
            return(TCL_OK);

        if ((entry = Tcl_FindHashEntry(&table,item)) == NULL) {
            // Dont complain about non-existent items 
            continue;
        }

        if (strncmp(item,"arg",3) == 0) {

            Cor_CArg *arg = (Cor_CArg *) Tcl_GetHashValue(entry);
            delete arg;

        } else if (strncmp(item,"tuple",5) == 0) {

            Cor_Tuple *tup = (Cor_Tuple *) Tcl_GetHashValue(entry);
            delete tup;

        } else {

            Tcl_AppendResult(interp,item,
                                ": can only delete arg or tuple items. ",0);
            err = 1;
            continue;
        }

        Tcl_DeleteHashEntry(entry);
    }

    if (err)
        return(TCL_ERROR);
    else
        return(TCL_OK);
}

// *************************************************************************
// Inputs: A scan_id some optional flag args, and an optional list. 
// Returns: An tuple_id (or {} if none remain) or "ok" if "-dispose"
//             is given. Or a list of arg_id's if "-args" is given. 
// Added By: Shaun Flisakowski Sept 6 1994
// Comments: The return value is passed in interp->result
// *************************************************************************

int Coral::GetTuple(Tcl_Interp *interp, int argc, char **argv)
{
    Tcl_HashEntry *entry;
    int args = 0, dispose = 0, i;
    char *scan_id;

    scan_id = *argv++;
    argc--;

    while(argc-- > 0) {

        if (strncmp(*argv,"-args",5) == 0)
            args = 1;
        else if (strncmp(*argv,"-dispose",8) == 0)
            dispose = 1;
        else { 
            Tcl_AppendResult(interp,"Coral: get_tuple: extra arguments.",0);
            return(TCL_ERROR);
        }

        argv++;
    }

    if ((entry = Tcl_FindHashEntry(&table,scan_id)) == NULL) {
        Tcl_AppendResult(interp,scan_id,": does not exist. ",0);
        return(TCL_ERROR);
    }
   
    Cor_ScanDesc *scan = (Cor_ScanDesc *) Tcl_GetHashValue(entry);

    if (scan->no_match())
        return(TCL_OK);

    Cor_Tuple *tuple = scan->next_tuple();

    if (scan->no_match())
        return(TCL_OK);

    if (args) {

        Tcl_DString list;

        Tcl_DStringInit(&list);

        int arity = tuple->arity();

        for (i=0; i < arity; i++){

            Cor_CArg *arg = new Cor_CArg((*tuple)[i]);
            Tcl_DStringAppendElement(&list, create_entry("arg",(void *) arg));
        }

        Tcl_DStringResult(interp, &list);
        // delete tuple;

    } else if (dispose) {

        Tcl_AppendResult(interp,"ok",0);
        // delete tuple;

    } else {

        Tcl_AppendResult(interp,create_entry("tuple",(void *) tuple),0);
    }

    return(TCL_OK);
}

// *************************************************************************
// Inputs: A arg_id pointer. 
// Returns: The type of the arg_id argument. 
// Added By: Shaun Flisakowski Sept 6 1994
// Comments: The return value is passed in interp->result
//                    types are:    int, float, string, functor, list
// *************************************************************************

int Coral::Identify(Tcl_Interp *interp, char *arg_id)
{
    Tcl_HashEntry *entry;

    if ((entry = Tcl_FindHashEntry(&table,arg_id)) == NULL) {
        Tcl_AppendResult(interp,arg_id,": does not exist. ",0);
        return(TCL_ERROR);
    }
   
    Cor_CArg *arg = (Cor_CArg *) Tcl_GetHashValue(entry);

    if (arg->is_empty()) {
        Tcl_AppendResult(interp,"empty",0);
        return(TCL_OK);
    }

    if (is_int(*arg)){
        Tcl_AppendResult(interp,"int",0);
        return(TCL_OK);
    }

    if (is_float(*arg)){
        Tcl_AppendResult(interp,"double",0);
        return(TCL_OK);
    }

    if (is_string(*arg)){
        Tcl_AppendResult(interp,"string",0);
        return(TCL_OK);
    }

    if (is_list(*arg)){
        Tcl_AppendResult(interp,"list",0);
        return(TCL_OK);
    }

    if (is_functor(*arg)){
        Tcl_AppendResult(interp,"functor",0);
        return(TCL_OK);
    }

    if (is_chunk(*arg)){
        Tcl_AppendResult(interp,"chunk",0);
        return(TCL_OK);
    }

    Tcl_AppendResult(interp,"Cannot identify argument",0);
    return(TCL_ERROR);
}

// *************************************************************************
// Inputs: A Coral query (as a string)
// Returns: A scan_id that can be used to get the matching tuples
//                through calls to get_tuple.
// Added By: Shaun Flisakowski Aug 10 1995
// Comments: The return value is passed in interp->result
//                    
//        Exs:  coral open "rel(_, _)"
//                    will open a scan of all tuples in relation "rel"
//                    
//        Exs:  coral open "rel(X, [_, X | _])"
//                    will open a scan of all tuples in relation "rel"
//                    whose first argument matches the second item in 
//                    list which is the second argument. 
//
// *************************************************************************

int Coral::OpenScan(Tcl_Interp *interp, int argc, char **argv)
{
    Tcl_HashEntry *entry;
    char *relname, *str_arity, *argval, *types;
    char *query;
    int   arity, i, m, vid=0, len;
    double r;
    Cor_ScanDesc *scan;

    query = *argv++;
    argc--;

    scan = open_scan(query);

    if (!scan){
        Tcl_AppendResult(interp,"cannot open scan on ", query,0);
        return(TCL_ERROR);
    }

    Tcl_AppendResult(interp,create_entry("scan",(void *) scan),0);
    return(TCL_OK);
}

// *************************************************************************
// Inputs: An arg_id (or tuple_id)
// Returns: A string equivilent of the arg (ideal for printing to
//                    a file or on the screen).
// Added By: Shaun Flisakowski Sept 6 1994
// Comments: The return value is passed in interp->result
// *************************************************************************

int Coral::sprint(Tcl_Interp *interp, char *arg_id)
{
    int pos = 0;
    Tcl_HashEntry *entry;

    if ((entry = Tcl_FindHashEntry(&table,arg_id)) == NULL) {
        Tcl_AppendResult(interp,arg_id,": does not exist. ",0);
        return(TCL_ERROR);
    }

    if (strncmp(arg_id,"arg",3) == 0) {

        Cor_CArg *arg = (Cor_CArg *) Tcl_GetHashValue(entry);
        arg->sprint(interp->result,&pos);

    } else if (strncmp(arg_id,"tuple",5) == 0) {

        Cor_Tuple *tup = (Cor_Tuple *) Tcl_GetHashValue(entry);
        tup->sprint(interp->result,&pos);

    } else {

        Tcl_AppendResult(interp, arg_id,
                ": can only print tuple or arg items. ",0);
        return(TCL_ERROR);
    }

    return(TCL_OK);
}

// *************************************************************************
// Inputs: An arg_id of a list.
// Returns: The coral list converted into a tcl list.
//             Returns a list of arg_id's if "-args" is given. 
// Added By: Shaun Flisakowski Sept 6 1994
// Comments: The return value is passed in interp->result
// *************************************************************************

int Coral::ConvertList(Tcl_Interp *interp, int argc, char **argv)
{
    int  pos, args = 0;
    char buff[200], *clst, *arg_id;
    Tcl_DString list;
    Tcl_HashEntry *entry;

    if (strncmp(*argv,"-args",5) == 0){
        argv++;
        argc--;
        args = 1;
    } 

    if (argc < 1) {
        Tcl_AppendResult(interp,"Coral: tcl_list: no arg_id given",0);
        return(TCL_ERROR);
    }

    arg_id = *argv;

    if ((entry = Tcl_FindHashEntry(&table,arg_id)) == NULL) {
        Tcl_AppendResult(interp,arg_id,": does not exist. ",0);
        return(TCL_ERROR);
    }
    
    Cor_CArg *arg = (Cor_CArg *) Tcl_GetHashValue(entry);

    if (! is_list(*arg)){
        Tcl_AppendResult(interp,arg_id,": is not a list.");
        return(TCL_ERROR);
    }

    Tcl_DStringInit(&list);

        /* Make each element of the coral list into a Tcl list element */

    Cor_CArg *rest = make_cdr(*arg);
    Cor_CArg *nxt_rest;

    while(rest && ! is_nil(*rest)){

        arg  = make_car(*rest);
        nxt_rest = make_cdr(*rest);
        delete rest;
        rest = nxt_rest;

        if (args){

            strcpy(buff,create_entry("arg",(void *) arg));

        } else {
            pos = 0;
            arg->sprint(buff,&pos);
        }

        clst = Tcl_DStringAppendElement(&list,buff);
    }

    Tcl_DStringResult(interp, &list);
    return(TCL_OK);
}

// *************************************************************************

Coral::~Coral()
{
    exit_coral();
}

void bomb()
{
    assert(0);
}

// *************************************************************************
//
// Inputs: A prefix and a void pointer to store.
// Outputs: The key to a hash table entry that stores the void pointer.
// Comments: The key is stored in a static area, it should be copied before
//                the next call to this function or it will be overwritten.  
// *************************************************************************

char *create_entry(char *prefix, void *store)
{
    static char buff[64];
    static int id = 0;            /* For making unique item names */
    int new_entry;
    Tcl_HashEntry *entry;

    do {
        sprintf(buff,"%s%d",prefix,id++);
        entry = Tcl_CreateHashEntry(&table,buff,&new_entry);
    } while(!new_entry);

    Tcl_SetHashValue(entry,store);
    return(buff);
}

// *************************************************************************
//
// Implementation of the tcl extended coral command -
// 
// usage: coral subcmd [options] 
// 
// sub commands:
// 
// 
// coral init ?program_name?
// coral exit
// coral consult file ?file?           Consult 1 or more files.
// coral open coral_query
// coral close scan_id
// coral get_tuple scan_id ?-dispose? ?-args? ?list?
// coral sprint tuple_id
// coral sprint arg_id
// coral arg tuple_id index           Returns arg_id
// coral delete arg_id ?arg_id...?
// coral delete tuple_id ?tuple_id...?
// coral ident arg_id                 Returns type of arg (one of: 
//                                         empty, int, double, string,
//                                         list, functor or chunk)
// coral tcl_list ?-args? arg_id      Returns coral list of arg_id as a tcl list
// coral parse string ?string?        Parse the string as coral input
//
// coral list_rels
//
// *************************************************************************

extern "C"
int coralCmd(ClientData cd, Tcl_Interp *interp, int argc, char **argv)
{
    int i, j, m, ret;
    Tcl_DString tmp;

    if (argc < 2) {
        Tcl_AppendResult(interp,"coral: no sub-command specified.",0);
        return(TCL_ERROR);
    }
    
    char *cmd = argv[1];
    int   len = strlen(cmd);

#ifdef     DEBUG
    cerr <<  "Coral command is: ";
    for (j=0; j<argc; j++){
        cerr << argv[j] << " ";
    }
    cerr << "\n\n";
#endif    /* DEBUG */
    
    if ((! coral) && (strncmp(cmd, "init", len) != 0)) {
        Tcl_AppendResult(interp, "Coral not initialized: \"",cmd,"\"", 0);
        return(TCL_ERROR);
    }


    switch (*cmd) {

    // ##################

    case 'a':

        if (strncmp(cmd, "arg", len) == 0){

            if (argc <= 3){
                Tcl_AppendResult(interp,"Usage: coral arg tuple_id index",0);
                return(TCL_ERROR);
            }

            return(coral->GetArg(interp,argv[2],argv[3]));
        }

        if (strncmp(cmd, "arity", len) == 0){
            if (argc == 4)
                m = Arity(argv[2], argv[3]);
            else
                m = 0;

#ifdef    DEBUG
            cerr << "The Arity of " << argv[2] << " " << argv[3]
                 << "  is: " << m << endl;
#endif
            return(TCL_OK);
        }
        break;

    // ##################

    case 'c':

        if (strncmp(cmd, "consult", len) == 0){

            if (argc <= 2){
                Tcl_AppendResult(interp,
                    "Coral: consult: no file name specified",0);
                return(TCL_ERROR);
            }

            argc -= 2;
            argv += 2;
            return(coral->Consult(interp,argc,argv));
        }

        if (strncmp(cmd, "close", len) == 0){

            if (argc <= 2){
                Tcl_AppendResult(interp,"Coral: close: no scan_id specified",0);
                return(TCL_ERROR);
            }

            return(coral->CloseScan(interp,argv[2]));
        }

        break;

    // ##################

    case 'd':

        if (strncmp(cmd, "delete", len) == 0){

            if (argc <= 2){
                Tcl_AppendResult(interp,"Coral: delete: no item to delete",0);
                return(TCL_ERROR);
            }

            argc -= 2;
            argv += 2;
            return(coral->DeleteItems(interp,argc,argv));
        }

        if (strncmp(cmd, "dump", len) == 0){

            if (argc <= 4){
                Tcl_AppendResult(interp,
                        "Usage: coral dump dest_rel src_rel arity",0);
                return(TCL_ERROR);
            }

            return(DumpRel(interp,argv[2],argv[3],argv[4]));
        }
        break;

    // ##################

    case 'e':

        if (strncmp(cmd, "ex_open", len) == 0){

            if (argc <= 3){
                Tcl_AppendResult(interp,
                    "Usage: coral ex_open relname pattern",0);
                return(TCL_ERROR);
            }

            return(RegExScan(interp,argv[2],argv[3]));
        }

        if (strncmp(cmd, "empty", len) == 0){

            if (argc <= 3){
                Tcl_AppendResult(interp,
                        "Usage: coral empty relname arity",0);
                return(TCL_ERROR);
            }

            return(EmptyRel(interp,argv[2],argv[3]));
        }

        if (strncmp(cmd,"exit",len)==0){
            if (coral != NULL) {
                delete coral;
                Tcl_DeleteHashTable(&table);
            }

            coral = (Coral *) NULL;
            return(TCL_OK);
        }
        break;

    // ##################

    case 'g':

        if (strncmp(cmd, "get_tuple", len) == 0){

            if (argc <= 2){
                Tcl_AppendResult(interp,
                    "Coral: get_tuple: no scan_id specified",0);
                return(TCL_ERROR);
            }

            argc -= 2;
            argv += 2;
            
            return(coral->GetTuple(interp,argc,argv));
        }

        break;

    // ##################

    case 'i':

        if (strncmp(cmd, "ident", len) == 0){

            if (argc <= 2){
                Tcl_AppendResult(interp,"Coral: ident: no arg_id specified",0);
                return(TCL_ERROR);
            }

            return(coral->Identify(interp,argv[2]));
            return(TCL_OK);
        }

        if (strncmp(cmd,"init",len)==0){
            if (coral != NULL){
                Tcl_AppendResult(interp, "coral already initialized", 0);
                return(TCL_ERROR);
            } else {
                if (argc <= 2)
                    coral = new Coral();
                else
                    coral = new Coral(argv[2]);

                glob_interp = interp;  // Save a pointer to this interpreter
                Tcl_InitHashTable(&table,TCL_STRING_KEYS);
            }

            return(TCL_OK);
        }
        break;

    // ##################

    case 'l':

        if (strncmp(cmd,"list_rels",len)==0){
            Cor_listRelsInDB(Cor_CurDB->name, &cerr);
            return(TCL_OK);
        }
        break;

    // ##################

    case 'o':

        if (strncmp(cmd, "open", len) == 0){

            if (argc <= 2){
                Tcl_AppendResult(interp,"Usage: coral open coral_query",0);
                return(TCL_ERROR);
            }

            argc -= 2;
            argv += 2;
            return(coral->OpenScan(interp,argc,argv));
        }
        break;

    // ##################

    case 'p':

        if (strncmp(cmd, "parse", len) == 0){

            if (argc <= 2){
                Tcl_AppendResult(interp,
                    "Coral: parse: no parse string specified.",0);
                return(TCL_ERROR);
            }

            argc -= 2;
            argv += 2;
            return(coral->Parse(interp,argc,argv));
        }
        break;

    // ##################

    case 's':

        if (strncmp(cmd, "sprint", len) == 0){

            if (argc <= 2){
                Tcl_AppendResult(interp,"Coral: sprint: ",
                            "no tuple_id or arg_id given",0);
                return(TCL_ERROR);
            }

            return(coral->sprint(interp,argv[2]));
        }
        break;

    // ##################

    case 't':

        if (strncmp(cmd, "tcl_list", len) == 0){

            if (argc <= 2){
                Tcl_AppendResult(interp,"Coral: tcl_list: no arg_id given",0);
                return(TCL_ERROR);
            }

            argc -= 2;
            argv += 2;
            return(coral->ConvertList(interp,argc,argv));
        }
        break;

    // ##################

    }

    Tcl_AppendResult(interp, "unknown coral subcommand: \"", cmd, "\"", 0);
    return(TCL_ERROR);
}

// *************************************************************************

