// *************************************************************************
//    CArg.i
//
//    The Coral Arg - All Arguments in Coral can (or will soon) be
//                    put inside of these.
//
//    Inlined functions
//
// *************************************************************************


  // If _CARG_C is defined and these items are not inlined, compile them here
#if (!defined(NO_INLINE_CARG)) || defined(_CARG_C)


#ifndef    _CARG_INLINE
#define    _CARG_INLINE

#include "CoralError.h"
#include "Entity.h"
#include "Debug.h"

// *************************************************************************
INLINE Cor_CArg::Cor_CArg(void)
{
    init();
}

// *************************************************************************
INLINE Cor_CArg::Cor_CArg(int n)
{
#ifdef    DO_PURE
    _doub  = 0;       // To stop purify from seeing UMR everywhere
    _hash = 0;
#endif
    my_arg_kind = INT_ARG;
    _int = n;
}

// *************************************************************************
INLINE Cor_CArg::Cor_CArg(double d)
{
    my_arg_kind = DOUBLE_ARG;
    _doub = d;
}

// *************************************************************************
INLINE Cor_CArg::Cor_CArg(CArgKind kind)
{
    init();
    my_arg_kind = kind;
}

// *************************************************************************
// Constructor for Expressions
INLINE Cor_CArg::Cor_CArg(CArgKind kind, Cor_CArg *larg, Cor_CArg *rarg)
{
    my_arg_kind = kind;

    expr._larg = larg;
    expr._rarg = rarg;

    _hash = UnknownHashValue;
}

// *************************************************************************
INLINE Cor_CArg::~Cor_CArg()
{
    free_mem();
}

// *************************************************************************
INLINE void Cor_CArg::init()
{
#ifdef    DEBUG
      // Just so these values aren't complete garbage
    _doub  = 0;
    _hash = 0;
#endif

    my_arg_kind = EMPTY_ARG;
}

// *************************************************************************
INLINE Cor_CArg& Cor_CArg::operator= (int i)
{
    my_arg_kind = INT_ARG;
    _int = i;
    return(*this);
}

// *************************************************************************
INLINE Cor_CArg& Cor_CArg::operator= (double d)
{
    my_arg_kind = DOUBLE_ARG;
    _doub = d;
    return(*this);
}

// *************************************************************************
INLINE int Cor_CArg::operator==(const Cor_CArg& c) const
{

    return( ((Cor_CArg*) this)->equals((Cor_CArg*) &c));
}

INLINE int Cor_CArg::equals(const Cor_CArg& c) const
{
    return( ((Cor_CArg*) this)->equals((Cor_CArg*) &c));
}

// *************************************************************************
INLINE void Cor_CArg::set_kind(CArgKind kind)
{
    my_arg_kind = kind;
}

// *************************************************************************
INLINE void Cor_CArg::set_num(int ival)
{
    my_arg_kind = INT_ARG;
    _int = ival;
}

INLINE void Cor_CArg::set_num(double dval)
{
    my_arg_kind = DOUBLE_ARG;
    _doub = dval;
}

// *************************************************************************
//  For the C++ Interface
INLINE void Cor_CArg::set_chunk(void *chnk)
{
#ifdef    DO_PURE
    _doub  = 0;
    _hash = 0;
#endif
    my_arg_kind = CHUNK_ARG;
    _chunk = chnk;
}

INLINE void Cor_CArg::set_var(int x)
{
    my_arg_kind = VAR_ARG;
    _var_num = x;
    ASSERT(_var_num >= 0);
}

// *************************************************************************
INLINE void Cor_CArg::set_grping(int arg_cnt)
{
extern Cor_CArg *Cor_NestedSymbol;

    my_arg_kind   = NO_AGG_GRPING_ARG;

    grp._nested_args  = new Cor_CArg(*Cor_NestedSymbol, arg_cnt);
    grp._seq_num      = 0;
    _chk_subsum       = 0;
}

INLINE void Cor_CArg::set_grping(Cor_CArg *nested_args,
        int seq_num, CArgKind kind)
{
    my_arg_kind   = kind;

    grp._nested_args  = nested_args;
    grp._seq_num      = seq_num;
    _chk_subsum       = 0;
}

// *************************************************************************
INLINE void Cor_CArg::set_grp_subsum(int chk_subsum)
{
    _chk_subsum = chk_subsum;
}

INLINE void Cor_CArg::set_grp_seqnum(int seqnum)
{
    grp._seq_num = seqnum;
}

INLINE void Cor_CArg::set_grp_kind(CArgKind kind)
{
    my_arg_kind   = kind;
}

// *************************************************************************
// This compares the symbolic form of items, they need not be the same type
INLINE int Cor_CArg::nm_equals(const Cor_CArg& arg) const
{
    Cor_CArg *mine = (Cor_CArg *) this;
    Cor_CArg *othr = (Cor_CArg *) &arg;

    switch (my_arg_kind) {

    default:
        break;

    case FUNCT_ARG:
        mine = func._functor;

    case SYMBOL_ARG:
    case TINY_SYM_ARG:

        switch (arg.my_arg_kind) {

        default:
            return(0);

        case FUNCT_ARG:
            othr = othr->func._functor;

        case SYMBOL_ARG:
        case TINY_SYM_ARG:
            // Need strcmp case not all types use string references
            if (strcmp(mine->symbol(),othr->symbol()) == 0)
                return(1);
            break;
        }
        break;
    }

    return(0);
}

// *************************************************************************
INLINE int Cor_CArg::equals(char *str) const
{
    ASSERT(is_symbol());

    switch (my_arg_kind) {

    case SYMBOL_ARG: 
        if (sym._symbol && (*str == *(sym._symbol))
                    && (strcmp(sym._symbol, str) == 0))
            return(1);
        break;

    case TINY_SYM_ARG:
        if ((*str == *_symbuf) && (strcmp(_symbuf, str) == 0))
            return(1);
        break;

    default:
        break;
    }

    return(0);
}

// *************************************************************************
INLINE int Cor_CArg::is_empty() const
{
    return(my_arg_kind == EMPTY_ARG);
}

// *************************************************************************
// no type checking
INLINE int Cor_CArg::compare_num(const Cor_CArg& arg2) const
{
    // Note that the odds of '==' even returning true for doubles
    // is pretty slim. This needs to be changed to a tolerance check.

    switch (my_arg_kind) {

    default:
        return(0);

    case INT_ARG:
        switch (arg2.my_arg_kind) {

        default:
            return(0);

        case INT_ARG:
            return (arg2._int < _int ? 1 : ((arg2._int == _int) ? 0 : -1));

        case DOUBLE_ARG:
            return ( arg2._doub < ((double) _int) ?
                   1 : ((arg2._doub == ((double) _int)) ? 0 : -1) );
        }

    case DOUBLE_ARG:

        switch (arg2.my_arg_kind) {

        default:
            return(0);

        case INT_ARG:
            return ( ((double)arg2._int) < _doub ?
                   1 : ((double)arg2._int) == _doub ? 0 : -1);

        case DOUBLE_ARG:
            return ( arg2._doub < _doub ?
                  1 : arg2._doub == _doub ? 0 : -1);
        }
    }
}

// *************************************************************************
INLINE int Cor_CArg::is_num() const
{
    return((my_arg_kind == INT_ARG) || (my_arg_kind == DOUBLE_ARG));
}

// *************************************************************************
INLINE double Cor_CArg::NumDval() const
{
    if (my_arg_kind == DOUBLE_ARG)
        return _doub;

    if (my_arg_kind == INT_ARG)
        return (double)_int;

    //CORAL_error(COR_FATAL_ERROR, "","NumDval");
    return(0.0);
}

// *************************************************************************
INLINE int Cor_CArg::NumIval() const
{
    if (my_arg_kind == INT_ARG)
        return _int;

    //CORAL_error(COR_FATAL_ERROR, "","NumIval");
    return(0);
}

// *************************************************************************
INLINE int Cor_CArg::compute_isConstant() const
{
    switch (_hash) {

    default:
        return(1);

    case VarHashValue:
        return(0);

        // This might not be a good idea, but we'll give it a try.
    case UnknownHashValue:
        hash(NULL); 
        return(_hash != VarHashValue);
    }

    // Unreachable
    return(0);
}

// *************************************************************************
INLINE int Cor_CArg::isConstant() const
{
    switch (my_arg_kind) {

    default:
#ifdef    DEBUG
        TELL;
        cerr << "Unknown kind in isConstant(): " << kind_str() << endl;
#endif
    case EMPTY_ARG:
        //CORAL_error(COR_FATAL_ERROR, "","isConstant");
        return(0);

    case INT_ARG:
    case DOUBLE_ARG:
    case SYMBOL_ARG:
    case TINY_SYM_ARG:
    case CHUNK_ARG:
        ASSERT(_hash != UnknownHashValue);
        return(1);

    case FUNCT_ARG:
        return(compute_isConstant());

    case VAR_ARG:
        return(0);

    case REL_ARG:
        return(1);

    case DONT_CARE_ARG:
    case PARAMETER_ARG:
        return(0);

      // Lists
    case LIST_ARG:
        return(compute_isConstant());

    case NIL_ARG:
        return(1);

      // Arithmetic Expressions
    case ADD_OP_ARG:
    case SUB_OP_ARG:
    case MULT_OP_ARG:
    case DIV_OP_ARG:
    case IDIV_OP_ARG:
    case MOD_OP_ARG:

    case LSHFT_OP_ARG:
    case RSHFT_OP_ARG:

    case BT_AND_OP_ARG:
    case BT_OR_OP_ARG:
    case BT_NOT_OP_ARG:
    case BT_XOR_OP_ARG:
        return(1);

      // The various types of Grouping Args
    case SUM_GRPING_ARG:
    case PROD_GRPING_ARG:
    case CNT_GRPING_ARG:
    case AVG_GRPING_ARG:

    case SUM_DGRPING_ARG:
    case PROD_DGRPING_ARG:
    case CNT_DGRPING_ARG:
    case AVG_DGRPING_ARG:

    case MIN_GRPING_ARG:
    case MAX_GRPING_ARG:
    case SET_GRPING_ARG:
    case ANY_GRPING_ARG:
    case NO_AGG_GRPING_ARG:
        return(1);

    case AGG_RESULT_ARG:
        return(1);
    }

    cerr << "Unimplemented cases in isConstant.\n";

    return(0);
}

// *************************************************************************

INLINE CArgKind Cor_CArg::Kind() const { return(my_arg_kind); }

INLINE int Cor_CArg::is_int()    const { return(my_arg_kind == INT_ARG); }
INLINE int Cor_CArg::is_short()  const { return(my_arg_kind == INT_ARG); }
INLINE int Cor_CArg::is_long()   const { return(my_arg_kind == INT_ARG); }
INLINE int Cor_CArg::is_float()  const { return(my_arg_kind == DOUBLE_ARG); }
INLINE int Cor_CArg::is_double() const { return(my_arg_kind == DOUBLE_ARG); }

INLINE int Cor_CArg::is_symbol() const {
    return((my_arg_kind == SYMBOL_ARG) || (my_arg_kind == TINY_SYM_ARG));
}
INLINE int Cor_CArg::is_string() const {
    return((my_arg_kind == SYMBOL_ARG) || (my_arg_kind == TINY_SYM_ARG));
}

INLINE int Cor_CArg::is_chunk()  const { return(my_arg_kind == CHUNK_ARG); }

INLINE int Cor_CArg::is_param()  const { return(my_arg_kind == PARAMETER_ARG); }

INLINE int Cor_CArg::is_rel()    const { return(my_arg_kind == REL_ARG); }

    // Dont Cares are converted into unique variables by the lexer
INLINE int Cor_CArg::is_dont_care() const
{
    return( (my_arg_kind == DONT_CARE_ARG)
            || (is_var() && nm_equals(*Cor_DontCareSymbol)) );
}

INLINE int Cor_CArg::is_list()   const { return(is_cons() || is_nil()); }
INLINE int Cor_CArg::is_cons()   const
{
    return(this && (my_arg_kind == LIST_ARG));
}

INLINE int Cor_CArg::is_nil()    const
{
    return((this == NULL) || (my_arg_kind == NIL_ARG));
}

INLINE int Cor_CArg::is_functor() const
{
    return(my_arg_kind == FUNCT_ARG);
}

INLINE int Cor_CArg::is_var()     const { return(my_arg_kind == VAR_ARG); }

INLINE int Cor_CArg::is_expr()    const
{
    return((my_arg_kind >= ADD_OP_ARG) && (my_arg_kind <= BT_XOR_OP_ARG));
}

INLINE int Cor_CArg::is_grouping() const
{
    return( (my_arg_kind >= SUM_GRPING_ARG)
            && (my_arg_kind <= NO_AGG_GRPING_ARG) );
}

INLINE int Cor_CArg::same_kind(Cor_CArg *carg) const
{
    if (my_arg_kind == carg->my_arg_kind)
        return(1);

    switch (my_arg_kind) {

    case INT_ARG:
        return(carg->my_arg_kind == DOUBLE_ARG);
    case DOUBLE_ARG:
        return(carg->my_arg_kind == INT_ARG);

    case TINY_SYM_ARG:
        return(carg->my_arg_kind == SYMBOL_ARG);
    case SYMBOL_ARG:
        return(carg->my_arg_kind == TINY_SYM_ARG);

    default:
        break;
    }

    return(0);
}

// *************************************************************************
INLINE int Cor_CArg::symlen() const
{
    switch (my_arg_kind) {

    default:
#ifdef   DEBUG
        cerr << "Attempting to call symlen() on CArg of kind "
             << kind_str() << endl;
#endif

    case EMPTY_ARG:
        return(0);

    case SYMBOL_ARG:
        if (sym._symbol)
            return(strlen(sym._symbol));
        return(0);

    case TINY_SYM_ARG:
        return(strlen(_symbuf));

    case FUNCT_ARG:
        return(strlen(functor()));
    }
}

INLINE char *Cor_CArg::symbol() const
{
    switch (my_arg_kind) {

    default:
#ifdef   DEBUG
        cerr << "Attempting to call symbol() on CArg of kind "
             << kind_str() << endl;
#endif
    case EMPTY_ARG:
        return(NULL);

    case SYMBOL_ARG:
        return(sym._symbol);

    case TINY_SYM_ARG:
        return((char *) _symbuf);

    case FUNCT_ARG:
        return(functor());
    }
}

INLINE HashVal Cor_CArg::symhash() const
{
#ifdef    DEBUG
    if (!is_symbol()){
        cerr << "Wrong type for symhash(), kind is " << kind_str() << endl;
    }
#endif

    ASSERT(is_symbol());

    return(_hash);
}

// *************************************************************************
INLINE void *Cor_CArg::chunk()
{
    return(_chunk);
}

// *************************************************************************

INLINE char *Cor_CArg::functor() const
{
    ASSERT(is_functor());

    switch (my_arg_kind) {

    default:
        ASSERT(0);
        return(NULL);

    case FUNCT_ARG:    
        return(func._functor->symbol());
    }
}

INLINE Cor_CArg *Cor_CArg::cfunctor() const
{
    ASSERT(is_functor());

    switch (my_arg_kind) {

    default:
        ASSERT(0);
        return(NULL);

    case FUNCT_ARG:    
        return(func._functor);
    }
}

INLINE Cor_ArgList& Cor_CArg::args()
{
    switch (my_arg_kind) {
    
    default:
        ASSERT(0);
    
    case FUNCT_ARG:    
        return(*(func._args));
    }
}

// *************************************************************************
INLINE int  Cor_CArg::var_num() const { return(_var_num); }

// *************************************************************************
INLINE Cor_Relation *Cor_CArg::rel() const
{
    ASSERT(my_arg_kind == REL_ARG);
    return(_rel);
}

// *************************************************************************
INLINE int Cor_CArg::num_items() 
{
    return(list._num_items);
}

INLINE int  Cor_CArg::incr_num_items(int chg)
{
    return(list._num_items += chg);
}

INLINE Cor_CArgStack *Cor_CArg::items()
{
    return(list._items);
}

// *************************************************************************
INLINE CArgKind  Cor_CArg::op() const
{
    ASSERT((my_arg_kind >= ADD_OP_ARG) && (my_arg_kind <= BT_XOR_OP_ARG));
    return(my_arg_kind);
}

INLINE Cor_CArg *Cor_CArg::larg() const
{
    ASSERT((my_arg_kind >= ADD_OP_ARG) && (my_arg_kind <= BT_XOR_OP_ARG));
    return(expr._larg);
}

INLINE Cor_CArg *Cor_CArg::rarg() const
{
    ASSERT((my_arg_kind >= ADD_OP_ARG) && (my_arg_kind <= BT_XOR_OP_ARG));
    return(expr._rarg);
}

// *************************************************************************
// Grouping
INLINE Cor_CArg *Cor_CArg::nested_args()
{
    return(grp._nested_args);
}

INLINE int Cor_CArg::grp_arity() const
{
    ASSERT((my_arg_kind >= SUM_GRPING_ARG)
            && (my_arg_kind <= NO_AGG_GRPING_ARG));

    if (grp._nested_args)
        return(grp._nested_args->arity());

    return(0);
}

// *************************************************************************
INLINE Cor_StorageRelation *Cor_CArg::dist_rel()
{
    return(agg._dist_rel);
}

// *************************************************************************

#ifdef    _CARG_C
#undef    _CARG_C
#endif

#endif    _CARG_INLINE
#endif    _CARG_C
