// *************************************************************************
//    BuiltinUnaryRelation.C
// *************************************************************************

#include "BuiltinUnaryRelation.h"

#include "InitBuiltins.h"
#include "TupleIterator.h"
#include "Globals.h"
#include "CoralError.h"
#include "Debug.h"

// *************************************************************************
Cor_BuiltinUnaryRelation::Cor_BuiltinUnaryRelation(int a, Cor_CArg *n,
        AggProc agg_op) : Cor_BuiltinRelation(a, n, NULL)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    op = agg_op;
}

// *************************************************************************
Cor_BindEnv *Cor_BuiltinUnaryRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BindEnv *env = iterator.bindenv;
    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match();
        return env;
    }
    iterator.stack_mark.get_mark();
    iterator.ipos++;

    if (iterator.arg_list.count() < 2) {
        *(Cor_exEnv->err_strm) << "Error: Not enough arguments provided.\n";
        return NULL;
    }

   // Term terms[2];
    for (int i = 0; i < 2; i++) {
        Cor_static_terms[i].expr = iterator.arg_list[i];
        Cor_static_terms[i].bindenv = env;
        Cor_full_dereference_term(Cor_static_terms[i]);
    }

    Cor_Relation *rel;
    switch (Cor_static_terms[0].Kind()) {

        case REL_ARG:
            rel = (Cor_static_terms[0].expr).rel();
            break;

        case FUNCT_ARG:
            rel = find_relation(Cor_static_terms[0].expr,
                    Cor_static_terms[0].expr.args().count());

            if (rel == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return NULL;
            }
            break;

         case SYMBOL_ARG:
            rel = find_relation(Cor_static_terms[0].expr, -1);
            if (rel == NULL) {
                CORAL_error(COR_REL_NOT_FOUND, NULL, __FUNCTION__);
                iterator.stack_mark.pop_to();
                return NULL;
            }
            break;

        default:
            CORAL_error(COR_BAD_ARG_TYPE, NULL, __FUNCTION__);
            *(Cor_exEnv->err_strm) << "Relation '" << name
                << "' has an invalid argument of type: "
                << Cor_static_terms[0].expr.kind_str() << endl;
            iterator.stack_mark.pop_to();
            return(NULL);
    }

    (*op)(Cor_usable_terms[0].expr, rel);
    Cor_usable_terms[0].bindenv = NULL;

    if (Cor_usable_terms[0].expr.is_empty()) {
        iterator.set_no_match();
        iterator.stack_mark.pop_to();
        return(env);
    }

    if ( ! Cor_unify_args (Cor_usable_terms[0], Cor_static_terms[1])) {
        iterator.set_no_match();
        iterator.stack_mark.pop_to();
    }

    return(env);
}

// *************************************************************************
