#include "CRdbManager.h"

#ifdef DO_RDB

#include "CRdbSolvers.h"
#include "SymTable.h"
#include "BuiltinRelation.h"
#include "Log.h"
#include "Rdb.h"
#include "Debug.h"



//---------------------------------------------------------------------------
//---------------------------------------------------------------------------

// Global Coral-Rdb handler -- set by coral_init, destroyed by exit_coral
Cor_CRdbManager *pCRdbManager = NULL;


//---------------------------------------------------------------------------

Cor_CRdbManager::Cor_CRdbManager()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    LOG( "Allocating rdb builtins" << endl );
    Cor_BuiltinRelation *temp;
    Cor_CArg carg;

    carg = Cor_CArg("rdb_open_db");
    temp = new Cor_BuiltinRelation(-1, &carg, RdbOpenDBSolver);

    carg = Cor_CArg("rdb_map");
    temp = new Cor_BuiltinRelation(-1, &carg, RdbMapSolver);

    carg = Cor_CArg("rdb_join");
    temp = new Cor_BuiltinRelation(-1, &carg, RdbJoinSolver);

    carg = Cor_CArg("rdb_commit");
    temp = new Cor_BuiltinRelation(-1, &carg, RdbCommitSolver);

    carg = Cor_CArg("rdb_rollback");
    temp = new Cor_BuiltinRelation(-1, &carg, RdbRollbackSolver);

    carg = Cor_CArg("rdb_execute");
    temp = new Cor_BuiltinRelation(-1, &carg, RdbExecuteSolver);

    carg = Cor_CArg("rdb_logging");
    temp = new Cor_BuiltinRelation(-1, &carg, RdbLoggingSolver);
}


//---------------------------------------------------------------------------

Cor_CRdbManager::~Cor_CRdbManager()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // remove any crdb relations still hanging around, before deleting db's
/*
    for( int i = crelList.Size() -1 ; i >= 0 ; i-- ) {
     	delete crelList[i];	// warning: this will call us back
		// to remove itself from the list
    }
    assert(crelList.Size() == 0 );
*/
}


//---------------------------------------------------------------------------

Cor_CRdbManager::Cor_DBEntry::Cor_DBEntry(const String newName,
        RdbDatabase *newDb)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    nameList.Append(newName);
    assert(newDb);
    db = newDb;
}


//---------------------------------------------------------------------------

Cor_CRdbManager::Cor_DBEntry::~Cor_DBEntry()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete db;
}


//---------------------------------------------------------------------------

RdbDatabase *Cor_CRdbManager::LookupDatabase(const String name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int size = dbList.Size();
    for( int i = 0 ; i < size ; i++ ) {
	Cor_DBEntry& dbe = *dbList[i];
	int numNames = dbe.nameList.Size();
	for( int j = 0 ; j < numNames ; j++ ) {
	    if( name == dbe.nameList[j] )
	      return dbe.db;
	}
    }
    return NULL;
}


//---------------------------------------------------------------------------

void Cor_CRdbManager::StoreDatabase(const String name,
				RdbDatabase *db)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int size = dbList.Size();
    for( int i = 0 ; i < size ; i++ ) {	// check for duplicates
	    Cor_DBEntry& dbe = *dbList[i];
	    if( dbe.db == db ) {
	        dbe.nameList.Append(name);
	        return;
    	}
    }
    // db has not been opened under another name.
    dbList.Append( new Cor_DBEntry(name, db) );
}


// inline void CRdbManager::StoreCRdbRelation(CRdbRelation *crel)
// {
//     crelList.Append(crel);
// }



//---------------------------------------------------------------------------

// void CRdbManager::RemoveCRdbRelation(CRdbRelation *crel)
// {
//     int s = crelList.Size();
//     for( int i = 0 ; i < s ; i++ ) {
// 	if( crelList[i] == crel )
// 	  crelList.Delete(i);
//     }
// }

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#endif // DO_RDB
