// **************************************************************************
//    main.C
// **************************************************************************

#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <iostream.h>

#include "CoralInit.h"
#include "Globals.h"
#include "ParseStack.h"
#include "CoralError.h"
#include "Util.h"
#include "BuiltinUtil.h"
#include "Collector.h"

#ifdef    DO_PVM
#include "PvmConnectArray.h"
#elif defined(DO_WODI)
#include "WodiConnectArray.h"
#else
#include "SockConnectArray.h"
#endif    // DO_PVM

#include "LocalConnectArray.h"
#include "ClientProtocol.h"
#include "Debug.h"

static char coral_version[] = "Coral Version 1.5.1";

// *************************************************************************

void show_sizes();
void Command_Line_Error(char *str);
int  GetFilePointers(char **src_str, int max_files);
int  GetAnswerStyle(char **src_str, int max_args);
void Command_Line_Error(char *str);

// *************************************************************************
int main(int ac, char **av)
{
    int i;
    unsigned short use_port = 0;

#ifdef    DO_CALL_STACK
    callstk = new CallStk();
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

#ifdef    DO_WODI
    av++;
    ac--;
#endif

      // We need these right away
    Cor_new_globals();

      // Need to know if we are in server mode ASAP
    for (i=1; i < ac; i++){
        if (strcmp(av[i],"-s") == 0)
            coralConfig->server_mode = 1;

        if (strncmp(av[i],"-p",2) == 0){
            if (strlen(av[i]) > 2)
                sscanf(&(av[i][2]),"%hu", &use_port);
        }
    }

    if (coralConfig->server_mode){

#ifdef    DO_PVM
        Cor_clients = new Cor_PvmConnectArray();
#elif defined(DO_WODI)
        cerr << "BEGIN OF CORAL SERVER ON WODI\n";
        cerr.flush();
        Cor_clients = new Cor_WodiConnectArray();
#else
          // Connect to the coral port
        if (use_port)
            Cor_clients = new Cor_SockConnectArray(use_port);
        else
            Cor_clients = new Cor_SockConnectArray();
#endif

        // Could not bind to port (or something, specific error message
        // given by lower level.)
        if (! Cor_clients) {
            cerr << "Error: Could not start Coral server.\n";
            exit(0);
        }
    } else {

          // Start up a "Local" Connection (use cin_strm/cout_strm)
        Cor_clients = new Cor_LocalConnectArray();
    }

    for (i=1; (i < ac) && (*av[i] == '-'); ++i) {

        switch (av[i][1]) {
        case 'a':
            // Answer style
            if (av[i][2] != '\0') Command_Line_Error(av[0]);
            i += GetAnswerStyle((char **)(av+i+1),(int)(ac - i -1));
            break;

        case 'c':
            // Read .coralrc mode ON : default
            Cor_exEnv->C_read_coralrc_default = 1;
            break;

        case 'C':
            // Read .coralrc mode OFF
            Cor_exEnv->C_read_coralrc_default = 0;
            break;

        case 'e':
            // Connect to EXODUS
            Cor_exEnv->C_connect_exodus_default = 1;
            break;

        case 'E':
            // Don't connect to EXODUS : default
            Cor_exEnv->C_connect_exodus_default = 0;
            break;

        case 'f':
        case 'F':
            // Input files
            if (av[i][2] != '\0')
                Command_Line_Error(av[0]);
            i += GetFilePointers((char **)(av+i+1), (int)(ac-i-1));
            break;

        case 'h':
        case 'H':
            // Print version and show usage 
            *cout_strm << "\nThis is " <<  coral_version << "\n\n";
            Command_Line_Error(av[0]);
            break;

        case 'i':
            // Interactive mode ON : default
            Cor_exEnv->C_interactive_mode_default = 1;
            break;

        case 'I':
            // Interactive mode OFF
            Cor_exEnv->C_interactive_mode_default = 0;
            break;


        case 'm':
            // Create .P.M mode ON : default
            Cor_exEnv->C_create_magic_default = 1;
            break;

        case 'M':
            // Create .P.M mode OFF
            Cor_exEnv->C_create_magic_default = 0;
            break;

        case 'o':
            // Output file
            if ((av[i][2] != '\0')||(i == ac-1))
                Command_Line_Error(av[0]);
            if (!set_out_strm(av[++i])) {
                *cerr_strm << av[0]
                    << " error: Cannot set print file to " << av[i] << endl;
                exit(1);
            }
            break;

        case 'p':
            // Display the prompt : default
            Cor_exEnv->C_print_prompt_default = 1;
            break;

        case 'P':
            // Don't display the prompt
            Cor_exEnv->C_print_prompt_default = 0;
            break;

        case 'q':
            // Quiet mode ON
            Cor_exEnv->C_quiet_mode_default = 1;
            break;

        case 'Q':
            // Quiet mode OFF : default
            Cor_exEnv->C_quiet_mode_default = 0;
            break;

        case 's':
            // server mode ON : not default;
            coralConfig->server_mode = 1;
            break;

        case 'v':
            // Verbose answer display ON : default
            Cor_exEnv->C_verbose_answer_default = 1;
            break;

        case 'V':
            // Verbose answer display OFF
            Cor_exEnv->C_verbose_answer_default = 0;
            break;

        case 'w':
            // Insert mode ON : default
            Cor_exEnv->C_insert_mode_default = 1;
            break;

        case 'W':
            // Insert mode OFF
            Cor_exEnv->C_insert_mode_default = 0;
            break;

        default:
            Command_Line_Error(av[0]);
        }
    }

    if (i<ac)
        Command_Line_Error(av[0]);

    Cor_exEnv->C_exec_mode = COR_DOING_INTERPRET;

      // If these things are compiled in, then set them up
    use_rdb = 1;
    use_exodus = 1;
    init_coral(av[0]);

#if    0
    Cor_CArg none;
    none.show_internals();
    show_sizes();
#endif

    if (coralConfig->server_mode){

          // Wait for a user
        Cor_clients->NextConn();
    }

    init_coral_interactive();

    if (!Cor_exEnv->C_quiet_mode_default && !coralConfig->server_mode) {
           // Print out the version of coral
        *(Cor_exEnv->out_strm) << "\nThis is " <<  coral_version << "\n";
          // Print preliminary message
        char name_buffer[1000] ;
        sprintf(name_buffer, "%s%s/init.msg", Cor_exEnv->coral_path, HELP_PATH);
        Cor_DisplayFile(name_buffer,0);
    }

    while (! Cor_parseStack->is_empty()){

        Cor_parseStack->parse(cerr_strm, 1,
                "syntax error -- skipping to next period\n");

        if (coralConfig->server_mode){

            if (!Cor_parseStack->is_empty()) {
                  // hack to fix server mode
                Cor_parseEnv->request_ack();
                Cor_parseEnv->print_prompt();
            }

            if (Cor_curConn && Cor_parseStack->is_empty()){

                  // Close the socket and delete the connection
                Cor_clients->DeleteConnection(Cor_curConn->conn_id);
            }

              // Switch users
#ifndef DO_WODI
            Cor_clients->NextConn();

              // Is current user part of a collection?
            while (Cor_curUser->collector) {

                  // Process the collection
                Cor_curUser->collector->ProcessUser(Cor_curUser);
                  // and try again.
                Cor_clients->NextConn();
            }
#endif // DO_WODI
        }
    }
    
    if (!Cor_exEnv->C_quiet_mode_default) {
        cerr << "\nHave a good day, " << cuserid(NULL) << "!\n";
    }

    exit_coral();

#ifdef    DO_CALL_STACK
    delete callstk;
    callstk = NULL;
#endif

    return(0);
}

// *************************************************************************
// Prints error message about command line error

void Command_Line_Error(char * str)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    cerr << "Input syntax : "
         << str << " [-o/O <output-file>][-f/F <input-file1> ...]\n";
    cerr << "      -a answer_style   Specify answer print "
         << "style (bindings, tuples, literal or empty)\n";
    cerr << 
      "      -c                Enable reading of .coralrc startup script\n";
    cerr << 
      "      -C                Disable reading of .coralrc startup script\n";
    cerr << "      -e                Connect to EXODUS\n";
    cerr << "      -E                Do not connect to EXODUS\n";
    cerr << "      -f <in_file>+     Specify input files\n";
    cerr << "      -h                Show usage (this message)\n";
    cerr << "      -i                Enable interactive mode\n";
    cerr << "      -I                Disable interactive mode\n";
    cerr << "      -m                Enable creation of.P.M files\n";
    cerr << "      -M                Disable creation of.P.M files\n";
    cerr << "      -o output_file    Specify output file\n";
    cerr << "      -p                Enable prompt printing\n";
    cerr << "      -P                Disable prompt printing\n";
    cerr << "      -q                Enable quiet mode\n";
    cerr << "      -Q                Disable quiet mode\n";
    cerr << "      -s                Enable server mode\n";
    cerr << "      -v                Enable verbose answer display\n";
    cerr << "      -V                Disable verbose answer display\n";
    cerr << "      -w                Enable insert mode\n";
    cerr << "      -W                Disable insert mode\n";

    cerr << "\nSpecial when in server mode:\n";
    cerr << "\n      -pNNNN          NNNN is the port to use\n";
    exit(1);
}

// *************************************************************************
// Reads the filenames specified in src_strin, opens the files, and saves
// the respective parsers onto the parseStack.  Also deletes the .M files
// corresponding to the opened files.
// *************************************************************************

int GetFilePointers(char **src_str, int max_files)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int count = 0;
    CoralStream *tmp_strm;
    char temp_str[200];
    
    if (coralConfig->server_mode){
        cerr << "File loading not supported in server mode\n";
        exit(0);
    }

      // Remove stdin from stack
    Cor_parseStack->pop();

    while ((count < max_files) && (*src_str[count] != '-')) {
        if ((tmp_strm = Cor_open_infile(src_str[count]))) {
            Cor_parseStack->put_on_bottom(tmp_strm,0,src_str[count]);
            sprintf(temp_str, "%s.M", src_str[count]);
            unlink(temp_str);
        } else {
            CORAL_error(COR_FILE_NOT_FOUND, src_str[count], "GetFileList",
                __FILE__, __LINE__);
        }
        count++;
    }

      // Put cin_strm back on the bottom
    Cor_parseStack->put_on_bottom(cin_strm,0,"stdin");
    Cor_parseStack->reset_position();

    return(count);
}

// *************************************************************************
// Reads the answer print style from the command line

int GetAnswerStyle(char **src_str, int max_args)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (max_args <= 0)
        return(0);

    if (!(strcmp(src_str[0],"tuple") && strcmp(src_str[0],"tuples")))
        Cor_exEnv->C_answer_style_default = COR_TUPLE_ANSWER_PRINT;
    else if (!(strcmp(src_str[0],"binding") && strcmp(src_str[0],"bindings")))
        Cor_exEnv->C_answer_style_default = COR_BINDING_ANSWER_PRINT;
    else if (!strcmp(src_str[0],"empty"))
        Cor_exEnv->C_answer_style_default = COR_EMPTY_ANSWER_PRINT;
    else if (!strcmp(src_str[0],"literal"))
        Cor_exEnv->C_answer_style_default = COR_LITERAL_ANSWER_PRINT;
    else
        Command_Line_Error(src_str[0]);

    return(1);
}

// *************************************************************************
