// *************************************************************************
//    GenericIndexSet.h
// *************************************************************************

#ifndef _GenericIndexSet_H_
#define _GenericIndexSet_H_

#include "Config.h"

#include "Context.h"
#include "BitVector.h"
#include "TupleIterator.h"
#include "GenericIndex.h"

// *************************************************************************

class Cor_StorageRelation;

// *************************************************************************

class Cor_GenericIndexSet {

  protected:
    Cor_Relation        *rel;
    int                  num_in_array;
    int                  max_in_array;
    Cor_GenericIndex   **Indices;

  public:
    Cor_GenericIndexSet(Cor_StorageRelation *reln) {
        rel = reln;
        num_in_array = 0;  // no indices available at initialization 
        max_in_array = COR_MAX_INDEXES;
        Indices = new Cor_GenericIndex*[COR_MAX_INDEXES];
    }

    virtual ~Cor_GenericIndexSet();

    int add_all_indices(Cor_StorageRelation *reln);
    // transfers its indices to reln.
    virtual int add_all_tuples(Cor_StorageRelation *, int, int ) { return 0;}
    virtual int add_all_goals(Cor_StorageRelation*, Cor_Context*,
            Cor_StorageRelation*) {
        return 0;
    }

    inline int num_indices() { return num_in_array; }

      // Delete this tuple from all indices
    virtual int  delete_tuple(Cor_Tuple *tuple) = 0;

    virtual Cor_GenericIndex *find_index( Cor_TupleIterator& ) = 0;
        // finds the index that best "matches" the given argument set in
        // the iterator
    virtual Cor_GenericIndex *find_index( Cor_Tuple*, int *tryindexp= NULL )=0;
    virtual Cor_GenericIndex *find_index( Cor_ArgList&, Cor_BindEnv*,
        int *tryindexp=NULL) = 0;

    inline Cor_GenericIndex*& operator[](int i) {return Indices[i];}

    virtual int insert_index( Cor_GenericIndex *Index ) = 0;
        // adds an index to the set of known indices

    virtual void fill_index(Cor_GenericIndex *Index) = 0;
        // fills in the index using default index in posn. 0

    virtual void print_adornments(ostream *outf);
};

// *************************************************************************

#define COR_I_FAIL             0        // failure for indices
#define COR_I_SUCCESS          1        // success for indices
#define COR_I_PARTIAL_SUCCESS  2

// *************************************************************************

#endif _GenericIndexSet_H_
