//**************************************************************************
//    WodiCoralClient.C
//        A Coral client using Wodi to compute Bill-of-material problem.
//**************************************************************************

#include <string.h>

#include "ClientProtocol.h"

// This is a hack but otherwise entire Coral library is linked in
#define  _NO_CORAL
#include "WodiStream.h"
#include "CoralStream.h"
#include <stdio.h>
#include <fstream.h>

#include "readline/readline.h" 
#include "readline/chardefs.h"
#include "readline/history.h"
#include "readline/keymaps.h"

//**************************************************************************

extern FILE  *rl_outstream;

extern "C" char *readline(char *);
extern "C" void add_history (char *);

#define MAX_WORKER_NUM 100

//**************************************************************************
// trim_right() -
//    Remove spaces tabs and newlines from the right side, check to make
//    sure the command is terminated with a space, if everything looks
//    okay, return 1.
//
//    Return: 1 if trimmed command appears complete, 0 if incomplete
//
//**************************************************************************
int trim_right(char *str)
{
    int j;
    int len = strlen(str);
    char *spt;

    for (j=len-1, spt=str+j; j >= 0; j--, spt--){

        if ((*spt == ' ') || (*spt == '\t') || (*spt == '\n')){
            *spt = '\0';
            len--;
        } else 
            break;
    }

    if (len && (str[len-1] == '.'))
        return(1);
    else
        return(0);
}

//**************************************************************************
int main(int argc, char * argv[])
{
#define w_out cout
#define w_err cerr
    
    // default
    char *consult_filename = "/afs/cs.wisc.edu/u/w/h/whuang/reef/wodi_test/test";
    char *output_filename = "/afs/cs.wisc.edu/u/w/h/whuang/reef/wodi_test/output";
    int num_machines = 1;
    int num_nodes = 100;
    
    // syntax:
    //     wodi_client -c <name of the consult file>
    //                 -o <name of the output file>
    //                 -n <number of machines required> (not implemented yet?)
    //                 -l <number of nodes for bom at level1>
    
    for (int l = 1; l < argc; l++) {
        if (!strncmp(argv[l], "-c",2))
            if (strlen(argv[l]) > 2)
                consult_filename = (argv[l] + 2);
        if (!strncmp(argv[l], "-o",2))
            if (strlen(argv[l]) > 2)
                output_filename = (argv[l] + 2);
        if (!strncmp(argv[l], "-n",2))
            sscanf(argv[l] + 2, "%u", &num_machines);
        if (!strncmp(argv[l], "-l",2))
            sscanf(argv[l] + 2, "%u", &num_nodes);
    }
    
    char Prompt[100];
    char Buf[2000];
    char *ReadBuf;
    char PassWord[12];

    int sync_cnt = 2;

    rl_outstream = stdout;

    strcpy(Prompt,"ready>> ");
    PassWord[0] = '\0';

    int  class_need, ac;
    int  class_needs[1];
    char * av[3];
        
    pvm_mytid();
    class_need = 1;
    class_needs[0] = num_machines;
    av[0] = "coral";
    av[1] = "-s";
    av[2] = NULL;

    wodi_init(0,NULL,NULL,class_need, class_needs, av,
              CORAL_WORK_TAG, CORAL_RESP_TAG, 0);
    
    char * query = new char[100];
    strcpy(query, "consult(\"");
    strcat(query, consult_filename);
    strcat(query, "\").");
    w_err << "Begin running the query: " << query << endl;

    Cor_wodibuf * pbuf[MAX_WORKER_NUM];
    CoralStream * cs[MAX_WORKER_NUM];
    for (l = 0; l < MAX_WORKER_NUM; l++) {
        pbuf[l] = new Cor_wodibuf(WODI_MASTER);
        cs[l] = new CoralStream(pbuf[l]);
    }
    
    CoralStream *in;
    CoralStream *out;

    FILE * file = fopen(output_filename, "w");

    double cost = 0;
    
    int step_num = 1;
    int cycle_num = 0;
    while (step_num <= num_nodes) {
        cycle_num++;
        for (l = 0; l < ((MAX_WORKER_NUM>num_nodes)?num_nodes:MAX_WORKER_NUM); l++) {
            
            pbuf[l]->wodi_id = step_num;

            time_t cal_time = time(NULL);
            char * s = new char[20];
            strftime(s, 20, "%m/%d %H:%M:%S",localtime(&cal_time));
            fprintf(file, "%s ", s);
            fprintf(file, "START STEP %d\n",step_num);
            fflush(file);
            
            out = cs[l];

            strcpy(query, "assign(answer_style, 1).\n");
            strcat(query, "consult(\"");
            strcat(query, consult_filename);
            char tmp[100];
            
            //each step actually consults a different file 
            //each file contains the rules defining BOM and data of some subtree
            sprintf(tmp,"%d", step_num);
            strcat(query, tmp);
            strcat(query, "\").\n");

            sprintf(tmp, "?bom(part0.0,X).\n",step_num-1);
            strcat(query, tmp);
            out->puts(query);
            out->printf("\nsync(%d).\n yield.\n",sync_cnt);
            out->flush();
            
            step_num++;
        }

        for (l = 0; l < ((MAX_WORKER_NUM>num_nodes)?num_nodes:MAX_WORKER_NUM); l++) {
            in = cs[l];

            int done = 0;
            int sync = 0;
            int ok_cnt = 2;

            int have_sent_yield = 0;

            do {
                Buf[4] = '\0';
                if (!(in->gets(Buf, 1000))) 
                    break;

                if (Buf[4] == ':'){

                    if (strncmp(Buf,CP_OKAY,5) == 0) {
                        if (sync)
                            if (--ok_cnt == 0)
                                done = 1;

                    } else if (strncmp(Buf,CP_SYNC,5) == 0) {
                        int val;
                        sscanf(&(Buf[5]), "%d", &val);
                        sync = 1;
                    } else if (strncmp(Buf,CP_CERR,5) == 0) {
                        w_out << (&(Buf[5]));

                    } else if (strncmp(Buf,CP_MESG,5) == 0) {
                        w_out << (&(Buf[5]));

                    } else if (strncmp(Buf,CP_LOGIN,5) == 0) {
                    } else if (strncmp(Buf,CP_USER,5) == 0) {
                        w_out << Buf;

                    } else if (strncmp(Buf,CP_PSWD,5) == 0) {
                        w_out << Buf;
                        // Save the password for use with Explain
                        strncpy(PassWord,&(Buf[5]),8);
                        PassWord[8] = '\0';
                        int i = strlen(PassWord);
                        if (PassWord[i-1] == '\n')
                            PassWord[i-1] = '\0';

                    } else if (strncmp(Buf,CP_R_ID,5) == 0) {
                        w_out << Buf;

                    } else if (strncmp(Buf,CP_PRMT,5) == 0) {

                        strcpy(Prompt,&(Buf[5]));
                        Prompt[strlen(Prompt)-1] = ' ';

                    } else if (strncmp(Buf,CP_EXIT,5) == 0) {

                        w_out << "exiting.\n";
                        return(0);

                    } else {
                        w_err << "Warning, unimplemented protocol header.\n";
                        w_out << Buf;
                    }
                } else {
                    w_out << Buf;
                    
                    // looking for patern (X) where X is a number
                    char * ptr = strchr(Buf, '(');
                    if (ptr) {
                        ptr++;
                        int i = 0;
                        while (isdigit(ptr[i]) || ptr[i] == '.' || ptr[i] == '+' || ptr[i] == 'e') 
                            i++;
                        if (ptr[i] == ')') {
                            ptr[i] = '\0';
                            cost += atof(ptr);
                            w_out << "cost += " << atof(ptr) << endl;
                        }
                    }                            
                }

            } while(!done);

            Buf[0] = '\0';

            time_t cal_time = time(NULL);
            char * s = new char[20];
            strftime(s, 20, "%m/%d %H:%M:%S",localtime(&cal_time));
            fprintf(file, "%s", s);
            fprintf(file, " STEP %d DONE\n", pbuf[l]->wodi_id);
            fflush(file);
            w_out.flush();
        }        
    }

    w_out << "\nAll Done!\n";

    w_err << "\nbye!\n";

    w_out << "cost = " << cost << endl;

    wodi_complete();
    return(0);


}
//**************************************************************************

