
%{
#include "Globals.h"
#include "Scanner.h"
#include "Profile.h"
#include "Annotations.h"
#include "Literal.h"
#include "Interface.h"
#include "ParseActions.h"
#include "CoralError.h"
#include "Execute.h"
#include "Database.h"
#include "Sips.h"
#include "Rule.h"
#include "Collection.h"

#include "ParseUtils.h"
#include "BuiltinSymbols.h"
#include "ArgList.h"
#include "CArg.h"
#include "ClientProtocol.h"

void yyerror(char *s);

EXTERN int yylex(YYSTYPE *lvalp);
EXTERN int yyparse();

static char *fname = "yyparse";

extern void COR_set_single_answer(Cor_ModuleInfo& Module, int val);

    // Redefine the parser to use this "lexer" so that we can switch
    // scanners (aka lexers) depending who is one the top of the stack.
    // The function will actually be called as:
    //     (parseEnv->lexer)(&yylval)
    // assuming YYPURE is defined and YYLSP_NEEDED is not.
    // This is done here to get around bison.simple's definition of YYLEX.
#undef    yylex
#define   yylex    (Cor_parseEnv->lexer)

Cor_SymTable *Cor_BuiltinRelTable;

%}

// This next line makes the parser re-entrant.
%pure_parser

%type   <pred>   pred clause_lhs clause_left clause_lhs_list clause_lhs_element
%type   <carg>    arg arg_fact arg_primary var_arg simple_arg
%type   <sip>     arc sip
%type   <sips>    sip_list sip_list_entry
%type   <carg>    id anno_name adorn_string opt_adorn_string
%type   <carg>    symbol relop
%type   <ca_list> arglist opt_arglist grouping_list
%type   <carg>    member_list
%type   <alias_list> literal_alias
%type   <rule_anno> rule_anno rule_anno_list opt_annotes sips_or_rule_anno
%type   <anno_case> anno_case anno_case_list
%type   <anno_case_name> anno_case_name
%type   <anno_case_details> anno_case_details
%type   <integer> plus_or_minus

// If a token is added here, it should be added to CArg.y as well.
// The parsers must have the same tokens because they are using the
// same lexers.
%token  CONS NIL
%token  APPEND UPDATE ASSIGN DELETE
%token  MARK LPAREN RPAREN COMMA COLON SLASHSLASH PERIOD LBRACE RBRACE ARROW
%token  SIPS LBRACKET RBRACKET UNDERSCORE VBAR PERCENT NOT QUESTION DOLLAR TRUE
%token  SUM PROD COUNT CMIN CMAX AVG ANY DISTINCT COUNT_D SUM_D PROD_D AVG_D
%token  DEL START_C_MODULE END_C_MODULE MODULE EXPORT ALIAS END_MODULE
%token  AGG_SELECTOR AGG_SELECTOR_MANY_TO_ONE AGG_SELECTOR_MANY_TO_MANY
%token  PLUS MINUS STAR SLASH EQUAL NOTEQUAL LESS GREATER LESSEQUAL GREATEREQUAL
%token  LSHIFT RSHIFT BITXOR BITAND BITNOT BITOR MOD
%token  AT CUT DOUB_AT
%token  <carg> FSYMBOL SYMBOL VAR OP RELOP
%token  <carg> NUM
%left   BITOR
%left   BITXOR
%left   BITAND
%left   LSHIFT RSHIFT
%left   PLUS MINUS
%left   STAR SLASH SLASHSLASH MOD
%right  BITNOT

%%
prog:
        {
            Cor_exEnv->C_interrupt_raised = 0;

            Cor_parserStruct->reset(1);

            if (Cor_parserStruct->cur_collect->count()) {
              Cor_parserStruct->all_rules->push(Cor_parserStruct->cur_collect);
              Cor_parserStruct->cur_collect = new Cor_Collection();
            }

            if (!coralConfig->server_mode)
                Cor_parseEnv->print_prompt();
        }
        clause_list
        {
        }
;

clause_list:    clause_list
        {

            Cor_parserStruct->rule_var_names.empty();
            Cor_parserStruct->rhspreds.empty();
            Cor_parserStruct->nested_index = 0;

            // Eliminate this soon - SPF
            if (Cor_exEnv->C_at_end_module) {
                  // IS THIS NECESSARY ?? - PRAVEEN
                  // This init is _still_ necessary for := (and
                  // probably other small cavets as well) - SPF
                  // Changed to reset
                Cor_parserStruct->reset(1);
                if (Cor_parserStruct->cur_collect->count()) {
                    Cor_parserStruct->all_rules->push(
                                Cor_parserStruct->cur_collect);
                    Cor_parserStruct->cur_collect = new Cor_Collection();
                }
            }
        }
        clause
        {
            if (Cor_exEnv->C_interrupt_raised != 0) {
                if (Cor_exEnv->C_interrupt_raised == 2) {

                    // Special Handling Interrupt
                    // used for quit() and yield()
                    // which require returning from yyparse()

                    if (Cor_exEnv->quit){
                          // quit().

                          // Inform the client of quit() if in server mode.
                        if (coralConfig->server_mode)
                            *(Cor_exEnv->out_strm) << endl << CP_EXIT
                                << endl << CP_OKAY << endl << flush;

                          // Remove everything from the stack
                        Cor_parseStack->abort_file();
                        while (! Cor_parseStack->is_empty())
                            Cor_parseStack->pop();

                    } else if (Cor_exEnv->yield){
                          // yield().
                        Cor_exEnv->yield = 0;
                    }
                    Cor_exEnv->C_interrupt_raised = 0;

                      // Return from yyparse()
                    YYACCEPT;
                }

                if (Cor_exEnv->C_interrupt_raised > 0) {
                   *(Cor_exEnv->err_strm) << 
                            "CORAL:: Interrupt occurred during execution!:";
                   *(Cor_exEnv->err_strm) << "resetting state !\n";
                   Cor_exEnv->err_strm->flush();
                    //abort a consult that is halfway thru
                   Cor_parseStack->abort_file();
                   Cor_parserStruct->reset(1);
                }
                   // else, the user does not want to see more answers
                   // to the query, and the value is -1
                Cor_exEnv->C_interrupt_raised = 0;
            }

            if (Cor_exEnv->C_do_history_default)
                Cor_parseEnv->handle_history();

            if (Cor_parseEnv->show_prompt)
                Cor_parseEnv->print_prompt();
#ifdef DEBUG
            if (Cor_T_Stack->count) {
                *(Cor_exEnv->err_strm)
                    << "Warning : non-zero tablestack = "
                    << Cor_T_Stack->count << "!\n";
                Cor_T_Stack->count = 0;
            }
#endif
        }
    |    /* Nothing */
;

clause_lhs:    id COLON clause_lhs
        {

            Cor_parserStruct->cur_rule->clause_names->add_to_front($1,0);
        }

    |    id MINUS COLON clause_lhs
        {

            Cor_parserStruct->cur_rule->clause_names->add_to_front($1,1);
        }

    |     clause_lhs_list
        {
        }
;

clause_lhs_list: clause_lhs_list COMMA clause_lhs_element

    |       clause_lhs_element
        {
        }
;

clause_lhs_element: symbol LPAREN  opt_arglist RPAREN
        {

            Cor_parserStruct->cur_rule->head = new Cor_Literal(Cor_parserStruct,
                        *($1), ($3)->count());

            if (Cor_parseEnv->complexRelation)
                Cor_parseEnv->complexRelation--;
            ($3)->build_arglist(Cor_parserStruct->cur_rule->head->args);
            delete $3;

            Cor_parserStruct->cur_rule->clause_names->empty();
            $1->destroy();
        }

    |     DEL symbol LPAREN opt_arglist RPAREN
        {
            Cor_Clause *cur_rule = Cor_parserStruct->cur_rule;
            Cor_Literal *lit = new Cor_Literal(Cor_parserStruct, *($2),
                       ($4)->count());

            if (Cor_parseEnv->complexRelation)
                Cor_parseEnv->complexRelation--;
            lit->negated = 1;
            ($4)->build_arglist(lit->args);
            delete $4;
            cur_rule->clause_names->empty();

            if (Cor_exEnv->in_query_loop && (cur_rule->head == NULL)) {
                cur_rule->head = lit;
            } else {
                if (cur_rule->head_deletes) {
                    Cor_Literal **tmp_head_dels = cur_rule->head_deletes;
                    cur_rule->head_deletes =
                        new Cor_Literal* [++(cur_rule->head_deletes_count)];
                    for (int j=0; j < cur_rule->head_deletes_count-1; j++)
                        (cur_rule->head_deletes)[j] = tmp_head_dels[j];

                    delete [] tmp_head_dels;
                } else {
                    cur_rule->head_deletes = new (Cor_Literal*)[1];
                    cur_rule->head_deletes_count++;
                }
                cur_rule->head_deletes[cur_rule->head_deletes_count-1]=lit;
            }
            COR_set_lazy_eval(*(Cor_parserStruct->CurModule), 0);
            $2->destroy();
        }

    |     symbol
        {
            Cor_parserStruct->cur_rule->head =
                    new Cor_Literal(Cor_parserStruct, *($1), 0);
            if (Cor_parseEnv->complexRelation)
                Cor_parseEnv->complexRelation--;
            Cor_parserStruct->cur_rule->clause_names->empty();
            $1->destroy();
        }
;

/* export pred1(adorn11, adorn12, ...), pred2(adorn21, ..), ... */
export_clause:    EXPORT arg_fact
        {
            Cor_AddExportInfo(Cor_parserStruct, $2);
            $2->destroy();
        }

    |    export_clause COMMA arg_fact
        {
            Cor_AddExportInfo(Cor_parserStruct, $3);
            $3->destroy();
        }

    |    EXPORT symbol LBRACKET opt_arglist RBRACKET
        {
            Cor_CArg *tmp = Cor_create_functor($4, $2, Cor_parserStruct);
            Cor_AddExportInfo(Cor_parserStruct, tmp);
            tmp->destroy();
            $2->destroy();
            delete $4;
        }

    |    export_clause COMMA symbol LBRACKET opt_arglist RBRACKET
        {
            Cor_CArg *tmp = Cor_create_functor($5, $3, Cor_parserStruct);
            Cor_AddExportInfo(Cor_parserStruct, tmp);
            tmp->destroy();
            $3->destroy();
            delete $5;
        }
;

clause_left:
        {
           Cor_parserStruct->lhs_flag = 1;

           Cor_parserStruct->cur_rule = new Cor_Clause();
           Cor_parserStruct->cur_rule->line_number
                   = Cor_parseEnv->cur_line_num;
        }
             clause_lhs
        {
           Cor_parserStruct->lhs_flag = 0;
           Cor_parseEnv->C_num_head_arith_lits =
                   Cor_parserStruct->rhspreds.length;
        }
;

clause:       clause_left clause_rhs
        {
            Cor_Clause *cur_rule = Cor_parserStruct->cur_rule;

            if (Cor_exEnv->in_query_loop) {
                Cor_parserStruct->assign_func(Cor_FactFunction);
                if (Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE)
                    YYACCEPT;
                Cor_parserStruct->execute();
                Cor_parserStruct->reset(1);
            } else {

                // assign a label to the rule if it has no label
              if (cur_rule->clause_names->is_empty()) {

                  cur_rule->clause_names->add(
                          Cor_parserStruct->label_counter++, 0);
              }

                // initialize the sn_rule_label_counter of the rule
              cur_rule->sn_rule_label_counter = 0;


                // transfer information from parserStruct to cur_rule
                // and insert cur_rule into the parserStruct's rulelist
              Cor_complete_rule(cur_rule, Cor_parserStruct);

                // So we only have one copy
              if (Cor_exEnv->C_exec_mode != COR_DOING_IMPERATIVE)
                  Cor_parserStruct->cur_rule = NULL;
            }
        }

    |    clause_left APPEND rhs_pred_list opt_annotes PERIOD
        {
            Cor_parserStruct->assign_func(Cor_AppendFunction);

            if (Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE) {
                // SPF - Add to collection and keep going
                YYACCEPT;
            } else {
                Cor_parserStruct->execute();

                  // Better not delete the head predicate.
                delete Cor_parserStruct->cur_rule;
                Cor_parserStruct->cur_rule = NULL;
            }
        }

    |       clause_left DELETE rhs_pred_list opt_annotes PERIOD
        {
            Cor_parserStruct->assign_func(Cor_DeleteFunction);

            if (Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE) {
                // SPF - Add to collection and keep going
                YYACCEPT;
            } else {
                Cor_parserStruct->execute();

                  // Better not delete the head predicate.
                delete Cor_parserStruct->cur_rule;
                Cor_parserStruct->cur_rule = NULL;
            }
        }

    |    clause_left UPDATE rhs_pred_list opt_annotes PERIOD
        {
          *(Cor_exEnv->err_strm) << "Sorry : Updates not implemented yet\n";
        }

    |    clause_left ASSIGN rhs_pred_list opt_annotes PERIOD
        {
            Cor_parserStruct->assign_func(Cor_AssignFunction);

            if (Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE) {
                // SPF - Add to collection and keep going
                YYACCEPT;
            } else {
                Cor_parserStruct->execute();

                  // Better not delete the head predicate.
                delete Cor_parserStruct->cur_rule;
                Cor_parserStruct->cur_rule = NULL;
            }
        }

    |    QUESTION
        {
              // LEAK - SPF
            // Should we indicate somehow that this is a query form?
            // Save old rule in a collection
            Cor_parserStruct->cur_rule = new Cor_Clause();
            Cor_parserStruct->cur_rule->line_number
                    = Cor_parseEnv->cur_line_num;
        }
                rhs_pred_list PERIOD
        {
            Cor_parserStruct->assign_func(Cor_QueryFunction);

            if (Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE) {
                // SPF - Add to collection and keep going
                YYACCEPT;
            } else {
                Cor_parserStruct->execute();
                Cor_parserStruct->reset(1);
            }
        }

    |       START_C_MODULE PERIOD
        {
            Cor_exEnv->c_module_separator = 0;
        }

    |       END_C_MODULE PERIOD
        {
            Cor_exEnv->c_module_separator = 1;
            if (Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE) {
                // SPF - here we will actually accept
                YYACCEPT;
            }
        }

    |       MODULE id PERIOD
        {
            Cor_parserStruct->reset(1);

                // SPF
            if (Cor_parserStruct->cur_collect->count()) {
               Cor_parserStruct->all_rules->push(Cor_parserStruct->cur_collect);
               Cor_parserStruct->cur_collect = new Cor_Collection();
            }

            Cor_exEnv->C_at_end_module = 0;
            Cor_exEnv->in_query_loop = 0;
            Cor_parserStruct->CurModule->name = *($2);
            strcat(Cor_parseEnv->prompt, "*");
            $2->destroy();
        }

    |       END_MODULE PERIOD
        {
            Cor_exEnv->C_at_end_module = 1;

            Cor_parserStruct->assign_func(Cor_EndModuleFunction);

            if (Cor_exEnv->C_exec_mode == COR_DOING_INTERPRET) {

               Cor_parserStruct->execute();

               Cor_parserStruct->all_rules->push(Cor_parserStruct->cur_collect);
               Cor_parserStruct->cur_collect = new Cor_Collection();

               Cor_parserStruct->reset(0);
               Cor_exEnv->in_query_loop = 1;

               if (Cor_CurDB->name.equals("default_db"))
                   Cor_parseEnv->prompt = coralConfig->prompt;
               else {
                   strcpy(Cor_parseEnv->prompt_name,
                        Cor_CurDB->name.symbol());
                   strcat(Cor_parseEnv->prompt_name, coralConfig->prompt);
               }

               Cor_parseEnv->print_prompt();

            } else if (Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE) {

                // SPF - actually accept here
                Cor_exEnv->in_query_loop = 1;
                YYACCEPT;

            } else {
                  // magic or factoring or existential optimization
                YYACCEPT;
            }
        }

    |    export_clause PERIOD

    |    AT symbol symbol opt_adorn_string plus_or_minus PERIOD
        {
            Cor_ArgList *al1 = new Cor_ArgList(1);

            if (!($4)->is_empty())
                (*al1)[0] = *($4);
            else
                al1->set_count(0);

            if (($2)->equals(*Cor_AllowedAdornSymbol)) {
                Cor_add_allowed_adornment(Cor_parserStruct, $3, al1 );

            } else if (($2)->equals(*Cor_MultisetSymbol)) {

                  // NOTE :: Currently this does not give multiset
                  // semantics. It merely acts as do_not_check_subsum.
                  // The user should not use this annotation in its
                  // current state.

                Cor_add_multiset_annotation(Cor_parserStruct, $3, al1, 0);

            } else if (($2)->equals(*Cor_CheckSubsumSymbol)) {
                Cor_add_multiset_annotation(Cor_parserStruct, $3, al1, $5);

            } else if (($2)->equals(*Cor_IndexDeltaSymbol)) {
                Cor_add_index_delta_annotation(Cor_parserStruct, $3, al1);

            } else if (($2)->equals(*Cor_ASyncRelSymbol)) {
                Cor_add_asyncrel_annotation(Cor_parserStruct, $3, al1);

            } else {
                CORAL_error(COR_BAD_ANNO, ($2)->symbol(),
                     "yyparse:anno3", __FILE__, __LINE__);
            }

            delete al1;
            $2->destroy();
            $3->destroy();
            $4->destroy();
        }

    |    AT symbol symbol opt_adorn_string LPAREN opt_arglist
            RPAREN LPAREN opt_arglist RPAREN plus_or_minus PERIOD
        {

            Cor_ArgList *al1 = new Cor_ArgList(($6)->count());
            ($6)->build_arglist(*al1);
            delete $6;

            Cor_ArgList *al2 = new Cor_ArgList(($9)->count());
            ($9)->build_arglist(*al2);
            delete $9;

            if (($2)->equals(*Cor_MakeIndexSymbol)) {
                Cor_add_make_index_annotation(
                  Cor_parserStruct->make_index_annotations, $3, $4, al1, al2);

            } else if (($2)->equals(*Cor_ChoiceSymbol)) {
                *(Cor_exEnv->err_strm)
                     << "ERROR:  choice no longer supported"
                     << " -- use aggregate selection `any' instead\n";

            } else {
                CORAL_error( COR_BAD_ANNO,($2)->symbol(),
                    "yyparse:anno1", __FILE__, __LINE__);
            }

            delete al1;
            delete al2;
            $2->destroy();
            $3->destroy();
            $4->destroy();
        }

    |    AT symbol symbol opt_adorn_string LPAREN opt_arglist
                    RPAREN plus_or_minus PERIOD
        {

            Cor_ArgList *al1 = new Cor_ArgList(($6)->count());
            ($6)->build_arglist(*al1);
            delete $6;

            if (($2)->equals(*Cor_MakeIndexSymbol)) {
              Cor_add_make_index_annotation(
                Cor_parserStruct->make_index_annotations, $3, $4, al1, NULL);

            } else if (($2)->equals(*Cor_DiskResidentSymbol)) {
                Cor_add_diskrel_annotation(Cor_parserStruct, $3, $4, al1);

            } else if (($2)->equals(*Cor_ASyncRelSymbol)) {
                Cor_add_asyncrel_annotation(Cor_parserStruct, $3, al1);

            } else {
                CORAL_error(COR_BAD_ANNO, ($2)->symbol(),
                        "yyparse:anno2", __FILE__, __LINE__);
            }

            delete al1;
            $2->destroy();
            $3->destroy();
            $4->destroy();
        }

    |    AT symbol symbol opt_adorn_string LPAREN opt_arglist RPAREN
                LPAREN opt_arglist RPAREN simple_arg plus_or_minus PERIOD
        {
            Cor_CArg *op_arg = $11;

            Cor_ArgList *al1 = new Cor_ArgList(($6)->count());
            ($6)->build_arglist(*al1);
            delete $6;

            Cor_ArgList *al2 = new Cor_ArgList(($9)->count());
            ($9)->build_arglist(*al2);
            delete $9;

            if (($2)->equals(*Cor_AggSelectionSymbol)) {

                Cor_add_agg_sel_annotation(Cor_parserStruct,
                        $3, $4, al1, al2, op_arg, AggSel_Regular);
                COR_set_lazy_eval(*(Cor_parserStruct->CurModule), 0);

            } else if (($2)->equals(*Cor_AggSel_PerIter_Symbol)) {

                Cor_add_agg_sel_annotation(Cor_parserStruct,
                        $3, $4, al1, al2, op_arg, AggSel_PerIter);
                COR_set_lazy_eval(*(Cor_parserStruct->CurModule), 0);

            } else if (($2)->equals(*Cor_AggSel_MultAnswers_Symbol)) {

                Cor_add_agg_sel_annotation(Cor_parserStruct,
                        $3, $4, al1, al2, op_arg, AggSel_MultAnswers);
                COR_set_lazy_eval(*(Cor_parserStruct->CurModule), 0);

            } else {
                CORAL_error(COR_BAD_ANNO, ($2)->symbol(),
                     "yyparse:anno4", __FILE__, __LINE__);
            }

            $2->destroy();
            $3->destroy();
            $4->destroy();
            $11->destroy();
            Cor_parserStruct->agg_flag = 0;
        }

    |       AT symbol plus_or_minus PERIOD
        {
            if (($2)->equals(*Cor_SupplementaryMagicSymbol))
                COR_set_supplementary_magic(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_OrdinaryMagicSymbol))
                COR_set_ordinary_magic(*(Cor_parserStruct->CurModule), $3);
            else if ((($2)->equals(*Cor_NoMagicSymbol)) && $3)
                COR_set_no_magic(*(Cor_parserStruct->CurModule));
            else if (($2)->equals(*Cor_UseFactoringSymbol))
                COR_set_use_factoring(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_UseOrdSearchSymbol))
                COR_set_use_ordsearch(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_LazyEvalSymbol))
                COR_set_lazy_eval(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_UsePipeliningSymbol))
                COR_set_pipelining(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_IndexDeltaSymbol))
                COR_set_index_deltas(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_MultisetSymbol))
                COR_set_multisets(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_CheckSubsumSymbol))
                COR_set_check_subsum(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_SingleSccSymbol))
                COR_set_single_scc(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_SingleAnswerSymbol))
                COR_set_single_answer(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_ConvertFunctionsSymbol))
                COR_set_convert_functions(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_MonotonicSymbol))
                COR_set_monotonic(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_NoRewritingSymbol))
                COR_set_rewriting(*(Cor_parserStruct->CurModule), !($3));
            else if (($2)->equals(*Cor_NoPreProcessingSymbol))
                COR_set_preprocessing(*(Cor_parserStruct->CurModule), !($3));
            else if (($2)->equals(*Cor_DiskResidentSymbol))
                COR_set_disk_resident(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_NestedSccEvalSymbol))
                COR_set_nested_scc_eval(*(Cor_parserStruct->CurModule), $3);
            else if (($2)->equals(*Cor_NaiveBacktrackSymbol))
                COR_set_naive_backtracking(*(Cor_parserStruct->CurModule), $3);
            else {
               *(Cor_exEnv->err_strm) << "Unknown annotation: '"
                     << ($2)->symbol() << "' ignored.\n";
            }
            $2->destroy();
        }

    |    PERIOD {}

    |    error PERIOD
        {

            CORAL_error(COR_BAD_SYNTAX, NULL, fname, __FILE__, __LINE__);
            yyerrok;
            Cor_parserStruct->rule_var_names.empty();
            Cor_parserStruct->rhspreds.length = 0;
            Cor_parserStruct->nested_index = 0;

              // might cause new errors, but we need to be able to parse
              // from the command line again.
            Cor_exEnv->C_at_end_module = 1;
            Cor_exEnv->in_query_loop = 1;

             // SPF - this rule need not be in the collection
            delete Cor_parserStruct->cur_rule;
            Cor_parserStruct->cur_rule = NULL;
            Cor_parseEnv->print_prompt();

            Cor_parseEnv->fix_history();
        }
        clause {}
;

plus_or_minus:
              PLUS      { $$ = 1;}
        |     MINUS     { $$ = 0; }
        |   /* None */  { $$ = 1;}
;

clause_rhs: MARK rhs_pred_list opt_annotes PERIOD
        {
        }
    |    PERIOD
        {
            Cor_parserStruct->cur_rule->annotations = NULL;
            Cor_parserStruct->cur_rule->sips        = NULL;
        }
;

opt_annotes:    ALIAS
        {
            Cor_parserStruct->cur_rule->sips = NULL;
        }

            literal_alias sips_or_rule_anno
        {
           $$ = NULL;
        }
    |      /* Nothing */
        {
            $$ = NULL;
            Cor_parserStruct->cur_rule->annotations = NULL;
            Cor_parserStruct->cur_rule->sips        = NULL;
        }
;

rhs_pred_list:    rhs_pred_list COMMA rhs_pred
    |    rhs_pred
;

rhs_pred:    pred
        {
            Cor_parserStruct->rhspreds.push((void *) $1);
        }
    |    /* Nothing */
;

pred:    symbol LPAREN opt_arglist RPAREN
        {

              // LEAK - SPF
            $$ = new Cor_Literal(Cor_parserStruct, *($1),
                      ($3)->count());

            ($3)->build_arglist(($$)->args);
            delete $3;
            $1->destroy();
        }

    |   symbol
        {
            $$ = new Cor_Literal(Cor_parserStruct, *($1), 0);
            $1->destroy();
        }

    |    arg relop arg
        {
            $$ = new Cor_Literal(Cor_parserStruct, *($2), 2);
            $$->arith_pred = 1;
            ($$->args)[0] = *($1);
            ($$->args)[1] = *($3);
            $1->destroy();
            //  Don't delete "relop" (aka $2)
            $3->destroy();
        }

    |    NOT arg relop arg
        {
            $$ = new Cor_Literal(Cor_parserStruct, *($3), 2);
            $$->negated = 1;
            $$->arith_pred = 1;
            ($$->args)[0] = *$2;
            ($$->args)[1] = *$4;
            $2->destroy();
            //  Don't delete "relop" (aka $3)
            $4->destroy();
        }

    |    NOT symbol LPAREN opt_arglist RPAREN
        {

            $$ = new Cor_Literal(Cor_parserStruct, *($2),
                      ($4)->count());

            $$->negated = 1;
            ($4)->build_arglist(($$)->args);
            delete $4;
            $2->destroy();
        }

    |    NOT symbol
        {
            $$ = new Cor_Literal(Cor_parserStruct, *($2), 0);
            $$->negated = 1;
            $2->destroy();
        }

    |       CUT
        {
            $$ = new Cor_Literal(Cor_parserStruct, *Cor_CutSymbol, 0);
        }
;

relop:   EQUAL         { $$ = Cor_EquOpSymbol; }
    |    NOTEQUAL      { $$ = Cor_NeqOpSymbol; }
    |    LESS          { $$ = Cor_LssOpSymbol; }
    |    GREATER       { $$ = Cor_GrtOpSymbol; }
    |    LESSEQUAL     { $$ = Cor_LeqOpSymbol; }
    |    GREATEREQUAL  { $$ = Cor_GeqOpSymbol; }
;

arglist:     arglist COMMA arg
        {

            if ( Cor_parserStruct->CurModule->ConvertFunctions
                    && ($3) && ($3)->is_expr() ) {

                  // need to create the equality literal
                Cor_CArgStack ca_list;

                ca_list.push($3);
                Cor_CArg *carg
                      = Cor_pull_out_arith_expr(Cor_parserStruct, &ca_list);
                ($1)->push(carg);
            } else {
                ($1)->push($3);
            }
        }

    |    arg
        {

            $$ = new Cor_CArgStack();
            ($$)->push($1);

            if ( Cor_parserStruct->CurModule->ConvertFunctions
                  && ($1) && ($1)->is_expr() ) {

                  // need to create the equality literal

                  // This function emptys out the CArgStack, but it
                  // was just allocated, so this is no problem. (otherwise
                  // you would use a new one, as in the previous action)
                Cor_CArg *carg = Cor_pull_out_arith_expr(Cor_parserStruct,$$);

                ($$)->push(carg);
            }
        }
;

grouping_list: arg
        {
            $$ = new Cor_CArgStack();
            ($$)->push($1);
        }
;

opt_arglist:    arglist
    |   /* Nothing */
        {
              // Actually allocate one so everyone can see its empty
            $$ = new Cor_CArgStack();
        }
;

arg:     arg_primary
    |    arg PLUS arg
        {

            $$ = new Cor_CArg(ADD_OP_ARG,$1,$3);
        }
    |    arg MINUS arg
        {

            $$ = new Cor_CArg(SUB_OP_ARG,$1,$3);
        }
    |    arg STAR arg
        {

            $$ = new Cor_CArg(MULT_OP_ARG,$1,$3);
        }
    |    arg SLASH arg
        {

            $$ = new Cor_CArg(DIV_OP_ARG,$1,$3);
        }
    |    arg SLASHSLASH arg
        {

            $$ = new Cor_CArg(IDIV_OP_ARG,$1,$3);
        }
    |       arg LSHIFT arg
        {

            $$ = new Cor_CArg(LSHFT_OP_ARG,$1,$3);
        }
    |       arg RSHIFT arg
        {

            $$ = new Cor_CArg(RSHFT_OP_ARG,$1,$3);
        }
    |       arg BITXOR arg
        {

            $$ = new Cor_CArg(BT_XOR_OP_ARG,$1,$3);
        }
    |       arg BITAND arg
        {

            $$ = new Cor_CArg(BT_AND_OP_ARG,$1,$3);
        }
    |       arg BITOR arg
        {
            $$ = new Cor_CArg(BT_OR_OP_ARG,$1,$3);
        }
    |     BITNOT arg_primary
        {
            $$ = new Cor_CArg(BT_NOT_OP_ARG,$2,NULL);
        }

    |       arg MOD arg
        {
            $$ = new Cor_CArg(MOD_OP_ARG,$1,$3);
        }
    |     MINUS arg_primary    %prec BITNOT
        {

            if (($2)->is_num()) {
                if (($2)->is_int())
                    $$ = new Cor_CArg(-(($2)->NumIval()));
                else
                    $$ = new Cor_CArg(-(($2)->NumDval()));
                $2->destroy();
            } else {
                $$ = new Cor_CArg(SUB_OP_ARG,$2,NULL);
            }
        }
;

arg_primary:    arg_fact
    |    LPAREN arg RPAREN
        {
            $$ = $2;
        }

simple_arg:   symbol LPAREN opt_arglist RPAREN
        {
            $$ = Cor_create_functor($3, $1, Cor_parserStruct);
            $1->destroy();
            delete $3;
        }

arg_fact:    symbol
    |    NUM
    |    var_arg
    |    UNDERSCORE
        {

            // NOTE : this is a kludge ! Unification with dont care terms
            // is tricky, so instead we force the DontCareSymbol to be
            // treated each time as a distinct variable. The parser
            // NEVER generates  TheDontCareArg !! The scanner NEVER
            // returns UNDERSCORE !! Instead it returns VAR for "_"
               //$$ = new Cor_CArg(DONT_CARE_ARG);
               $$ = (Cor_CArg *) NULL;
               *(Cor_exEnv->err_strm) << "DONT_CARE_ARG!!\n";
        }

    |       DOLLAR
        {
            $$ = new Cor_CArg(PARAMETER_ARG);
        }

    |    simple_arg
    |    LBRACKET arglist VBAR arg RBRACKET
        {
            $$  = Cor_create_list($2, $4);
            delete $2;
        }
    |    LBRACKET opt_arglist RBRACKET
        {
            $$ = Cor_create_list($2, NULL);
            delete $2;
        }
    |    CONS LPAREN arg COMMA arg RPAREN
        {
            $$ = new Cor_CArg($3,$5);
            //$3->destroy();
            //$5->destroy();
        }
    |    NIL
        {
            $$ = new Cor_CArg(NIL_ARG);
        }
    |    LESS grouping_list GREATER
        {

              // Temporially disabled, until find spot to set agg_flag
            if (1 || Cor_parserStruct->lhs_flag || Cor_parserStruct->agg_flag) {
                Cor_parseEnv->complexRelation++;
                if (Cor_exEnv->dbg_aggregates)
                    *(Cor_exEnv->err_strm)
                        << "nested att found complex relation "
                        << Cor_parseEnv->complexRelation << endl;
                $$ = Cor_create_multi_grouping(Cor_parserStruct,$2);
            } else {
                CORAL_error(COR_GROUPING_ON_RHS, NULL, fname, __FILE__,
                        __LINE__);
                YYERROR;
            }
            delete $2;
        }

    |   LBRACE member_list RBRACE
        {

            $$ = $2;

            Cor_parseEnv->complexRelation++;

            if (Cor_exEnv->dbg_aggregates)
                *(Cor_exEnv->err_strm)
                    << "nested att found complex relation "
                    << Cor_parseEnv->complexRelation << endl;
        }
;

var_arg: VAR
        {
            
            int vn =
                Cor_varname_enter(&(Cor_parserStruct->rule_var_names), $1);
            $$ = new Cor_CArg();
            $$->set_var(vn);
        }

member_list:    member_list COMMA arg_fact
        {
            Cor_insert_into_set($1, $3);
            $3->destroy();
        }

    |       arg_fact
        {

            $$ = Cor_create_nested_multiset();
            Cor_insert_into_set($$, $1);
            $1->destroy();
        }
    |       /* Nothing - allows the empty set {} */
        {
            $$ = Cor_create_nested_multiset();
        }
;

literal_alias:    LBRACKET
        {
            Cor_parserStruct->sip_var_names.empty();
        }
            p_idlist RBRACKET
        {
            $$ = NULL;
        }
;

p_id:        id
        {
            Cor_parserStruct->sip_var_names.push($1);
           // can't  delete $1, see Array.C, add only copies ptr!
           // $1->destroy();
        }
;

p_idlist:    p_idlist COMMA p_id
    |     p_id
    |    /* Nothing */
;

sips_or_rule_anno: rule_anno_list
        {
            $$ = NULL;
        }
;

rule_anno_list:    rule_anno_list COMMA rule_anno
        {

            if ($3 != NULL)
                $$ = Cor_link_rule_anno($3,$1);
            else
                $$ = $1;
        }
    |    rule_anno
        {
            $$ = Cor_link_rule_anno($1,NULL);
        }
    |    /* Nothing */
        {
            $$ = NULL;
        }
;

rule_anno:    anno_name anno_case_details
        {
            $$ = Cor_make_anno($1,$2);
        }
    |     sips
        {
            $$ = NULL;
        }
;

anno_name:    symbol
;

anno_case:    anno_case_name ARROW anno_case_details
        {
            $$ = Cor_make_anno_case($1,$3);
        }
;

anno_case_list:    anno_case COMMA anno_case_list
        {
            $$ = Cor_link_anno_case_list($3, $1);
        }
    |     anno_case
        {
            $$ = Cor_link_anno_case_list($1,NULL);
        }
;

anno_case_name:      symbol
        {
            $$ = Cor_make_anno_case_name($1);
        }
    |    NUM
        {
            $$ = Cor_make_anno_case_name($1);
        }

    |    VAR
        {
            int vn =
                Cor_varname_enter(&(Cor_parserStruct->rule_var_names), $1);
            Cor_CArg *carg = new Cor_CArg();
            carg->set_var(vn);
            $$ = Cor_make_anno_case_name(carg);
        }

    |     UNDERSCORE
        {
            Cor_CArg *tmp = new Cor_CArg(DONT_CARE_ARG);
            $$ = Cor_make_anno_case_name(tmp);
        }

    |    symbol LPAREN opt_arglist RPAREN
        {
            $$ = Cor_make_anno_case_name(
                    Cor_create_functor($3, $1, Cor_parserStruct));
            $1->destroy();
            delete $3;
        }

    |     LBRACKET opt_arglist RBRACKET
        {
            $$ = Cor_make_anno_case_name(Cor_create_list($2, NULL));
            delete $2;
        }
;

anno_case_details:    symbol
        {
            $$ = Cor_make_anno_case_details($1);
        }

    |    NUM
        {
            $$ = Cor_make_anno_case_details($1);
        }

    |    VAR
        {
            int vn =
                Cor_varname_enter(&(Cor_parserStruct->rule_var_names), $1);
            Cor_CArg *carg = new Cor_CArg();
            carg->set_var(vn);
            $$ = Cor_make_anno_case_details(carg);
        }

    |    UNDERSCORE
        {
            Cor_CArg *tmp = new Cor_CArg(DONT_CARE_ARG);
            $$ = Cor_make_anno_case_details(tmp);
        }

    |    symbol LPAREN opt_arglist RPAREN
        {
            $$ = Cor_make_anno_case_details(
                    Cor_create_functor($3, $1, Cor_parserStruct));
            $1->destroy();
            delete $3;
        }

    |     LBRACKET opt_arglist RBRACKET
        {
            $$ = Cor_make_anno_case_details(
                    Cor_create_list($2, NULL));
            delete $2;
        }

    |    LBRACKET  anno_case_list RBRACKET
        {
            $$ = Cor_make_anno_case_details($2);
        }
;

sips:    SIPS LBRACKET sip_list RBRACKET
        {
            Cor_parserStruct->cur_rule->sips = $3;
        }
;


sip_list_entry: LBRACKET adorn_string
                /** LBRACKET ad_idlist RBRACKET  **/
            COLON sip RBRACKET
        {
            int i;
            char *ap;

            $$ = new Cor_Sips();

              // ??? Check for possible length errors
            $$->adorn.setlen(($2)->symlen());
            for(i=0,ap = ($2)->symbol();
                            i < ($2)->symlen(); i++, ap++)
                if ( *ap == 'b')
                    $$->adorn.set(i);

            $$->sip_list = $4;
            $$->next = NULL;
            delete $2;
        }
;

sip_list:    sip_list COMMA sip_list_entry
        {
            $$ = $3;
            $3->next = $1;

        }
    |    sip_list_entry
        {
            $$ = $1;
            $1->next = NULL;
        }

    |    LBRACKET RBRACKET
        {
            $$ = NULL;
        }
;

opt_adorn_string:    /* Nothing */
        {
            $$ = NULL;
        }
    |    LBRACKET id RBRACKET
        {
            $$ = $2;
        }
;

adorn_string:    id
;

sip:        sip COMMA arc
        {
            $$ = $1;
            $$->next = $3;
        }
    |    arc
;

arc:    LBRACKET arc_idlist RBRACKET
                ARROW id LPAREN bound_var_list RPAREN
        {

            $$ = new Cor_Sip();
            $$->tail = Cor_parseEnv->CurTail;
            $$->bound_vars = Cor_parseEnv->CurBoundVars;
            $$->head =
                Cor_varname_lookup(&(Cor_parserStruct->sip_var_names), $5);
            $$->next = NULL;
            Cor_parseEnv->CurTail.clear();
            Cor_parseEnv->CurBoundVars.clear();
            $5->destroy();
        }

    |    /* Nothing */
        {
            $$ = NULL;
        }
;

bound_var:    VAR
        {

            Cor_parseEnv->CurBoundVars.set(
                Cor_varname_lookup(&(Cor_parserStruct->rule_var_names), $1));
            $1->destroy();
        }
    |   /* Nothing */
;

bound_var_list:    bound_var_list COMMA bound_var
    |    bound_var
;

arc_idlist:    arc_idlist COMMA arc_id
    |    arc_id
;

arc_id:     id
        {
            Cor_parseEnv->CurTail.set(
                Cor_varname_lookup(&(Cor_parserStruct->sip_var_names), $1));
            $1->destroy();
        }
;

id:      symbol
    |    VAR
;

symbol:  SYMBOL
    |    FSYMBOL
;

%%

// *************************************************************************
void yyerror(char *s)
{
    *(Cor_exEnv->err_strm) << s << " line "
            << Cor_parseEnv->cur_line_num << ":\n"
            << Cor_parseEnv->commandbuf << endl;

    if (Cor_exEnv->C_interrupt_raised > 0) {

        *(Cor_exEnv->err_strm)
               << "CORAL:: Interrupt occurred during execution!:";
        *(Cor_exEnv->err_strm) << "resetting state !\n";
        *(Cor_exEnv->err_strm) << "Type a period to continue.\n" << flush;

          // abort a consult that is halfway thru
        Cor_parseStack->abort_file();
        Cor_parserStruct->reset(1);
        Cor_exEnv->C_interrupt_raised = 0;
    }
}

// *************************************************************************
