// *************************************************************************
//    ParseUtils.C
// *************************************************************************

#include <stdio.h>
#include <string.h>

#include "ParseUtils.h"
#include "ParserStruct.h"
#include "ArgList.h"
#include "Globals.h"
#include "HashSimpleRelation.h"
#include "BuiltinSymbols.h"
#include "Stack.h"
#include "Literal.h"
#include "ExportInfo.h"
#include "Rule.h"
#include "Util.h"
#include "Collection.h"
#include "Debug.h"


// *************************************************************************

static int check_builtin_op(Cor_CArg& name);
static int is_convertible_pred(Cor_CArg& pred_name);
static void simple_reorder_arith_preds(Cor_ParserStruct *Cor_parserStruct);

// *************************************************************************
// Allocate a new relation with subsumption turned off to store a nested
// multiset.
Cor_CArg *Cor_create_nested_multiset()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_StorageRelation *rel = new Cor_HashSimpleRelation(NESTED_SET_ARITY, 0,
        COR_DEFAULT_SET_BUCKETS);

    rel->set_subsumption(0);
    rel->name = *Cor_NestedSymbol;

    Cor_CArg *crel = new Cor_CArg(rel);

    return(crel);
}

// *************************************************************************
//    Insert an element into a nested set.
void Cor_insert_into_set(Cor_CArg *set, Cor_CArg *element)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ArgList *copy_args = new Cor_ArgList(NESTED_SET_ARITY); 

    (*copy_args)[NESTED_SET_ARITY-1] = *element;
    Cor_Tuple *newtuple = new Cor_Tuple(copy_args);
    set->rel()->insert_tuple(newtuple);
}

// **************************************************************************
Cor_CArg *Cor_create_list(Cor_CArgStack *ca_list, Cor_CArg *cdr)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (cdr && cdr->is_cons()) {

        while (ca_list->count() != 0) {
            Cor_CArg *car = ca_list->pop();
            cdr->items()->push(car);
            cdr->incr_num_items();
        }

    } else {
        if (!cdr)
            cdr = new Cor_CArg(NIL_ARG);

        if ((ca_list->count() != 0) || !cdr->is_list()) {
            Cor_CArgStack *stk = new Cor_CArgStack();
            stk->push(cdr);

            Cor_CArg *list = new Cor_CArg(stk);

            cdr = Cor_create_list(ca_list, list);
        }
    }

    if (cdr->is_cons())
        ASSERT(cdr->num_items() > 1);

    return(cdr);
}

// *************************************************************************
int Cor_varname_enter(Cor_CArgStack *var_names, Cor_CArg *s)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    // NOTE: this is a kludge! Unification with dont care terms is tricky,
    // so instead we force the DontCareSymbol to be treated each time as a
    // distinct variable.  The parser NEVER generates TheDontCareArg !!

    if (! s->equals(*Cor_DontCareSymbol)) {
        for (i=0; i < var_names->length; ++i) {
            if (s->equals(*(var_names->get_ith(i)))) {
                delete s;
                return(i);
            }
        }
    }

    var_names->push(s);
    return(var_names->length - 1);
}

// *************************************************************************
int Cor_varname_lookup(Cor_CArgStack *var_names, Cor_CArg *s)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    if (!s) 
      return (0);

    for (i=0; i < var_names->length; ++i) {
        if (s->equals(*(var_names->get_ith(i))))
            return(i);
    }

    *(Cor_exEnv->err_strm) << "Error: line "
        << Cor_parseEnv->cur_line_num << ": "
        << s->symbol() << " not in rule\n";

    return(0);
}

// *************************************************************************
static void simple_reorder_arith_preds(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_VoidStack   tempArray;

    int    curPred1 = 0;     // Current predicate in parserStruct.rhspreds

    if (Cor_parseEnv->C_num_head_arith_lits==0) 
        return;        // No simple reordering to be done 

    for (curPred1 = Cor_parseEnv->C_num_head_arith_lits; 
            curPred1 < Cor_parserStruct->rhspreds.length; curPred1++) {
        tempArray.push((void *)(Cor_parserStruct->rhspreds.get_ith(curPred1)));
    }

    for (curPred1 = 0; curPred1 < Cor_parseEnv->C_num_head_arith_lits;
                            curPred1++) {
        tempArray.push((void *)(Cor_parserStruct->rhspreds.get_ith(curPred1)));
    }

    // NOTE: The memory allocated for one of the arrays must be freed
    //       This is not done as of now!!

        // Well, lets try it
    delete [] Cor_parserStruct->rhspreds.data;
    
    Cor_parserStruct->rhspreds.data = tempArray.data;
    Cor_parserStruct->rhspreds.length = tempArray.length;
    Cor_parserStruct->rhspreds.data_size = tempArray.data_size;

      // To keep destructor of tempArray from destroying array 
    tempArray.data = NULL;
}

// *************************************************************************
// Notice the parserStruct argument is optional
Cor_CArg *Cor_create_functor(Cor_CArgStack *ca_list, Cor_CArg *op,
        Cor_ParserStruct *parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // These need to match the ordering in CArg.h and Gram.y
#define  FIRST_GRPING    ((int) SUM_GRPING_ARG)
#define  LAST_GRPING     ((int) NO_AGG_GRPING_ARG)
#define  MAX_GRPING      (LAST_GRPING - FIRST_GRPING +1)

    int builtin_index = check_builtin_op(*op);

      // Check if something like: sum(<ARG>).
    if ((builtin_index >= 0) && (builtin_index < MAX_GRPING)
            && (ca_list->count() == 1)) {

        Cor_CArg *carg = ca_list->pop();

        if (carg->Kind() == NO_AGG_GRPING_ARG) {
            CArgKind kind;

            kind = (CArgKind) (builtin_index + FIRST_GRPING);
            carg->set_kind(kind);
            return(carg);
        }

        // Item must be of the form max(C), process as a functor
        ca_list->push(carg);
    }

    if (parserStruct && parserStruct->CurModule->ConvertFunctions
            && is_convertible_pred(*op)) {

          // Rewrite: P(x op y) => op(x, y, TMP$i), P(TMP$i)
        Cor_Literal *pred;
        char tmp_buf[20];

        sprintf(tmp_buf, "TMP_%d_", Cor_parserStruct->rhspreds.length);

        Cor_CArg *carg = new Cor_CArg(tmp_buf);
        int vn = Cor_varname_enter(&(Cor_parserStruct->rule_var_names), carg);
        carg = new Cor_CArg();
        carg->set_var(vn);

        ca_list->push(carg);

        pred = new Cor_Literal(Cor_parserStruct, *op,
                ca_list->count());

        ca_list->build_arglist(pred->args);

        Cor_parserStruct->rhspreds.push((void *) pred);

        return(carg);
    }

      // count can be zero sometimes
    Cor_CArg *farg = new Cor_CArg(*op, ca_list->count());
    ca_list->build_arglist(farg->args());

    return(farg);

#undef  FIRST_GRPING
#undef  LAST_GRPING
#undef  MAX_GRPING
}

// *************************************************************************
Cor_CArg *Cor_pull_out_arith_expr(Cor_ParserStruct *Cor_parserStruct,
        Cor_CArgStack *ca_list)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *pred;
    char tmp_buf[20]; 

    sprintf(tmp_buf, "TMP_%d_", Cor_parserStruct->rhspreds.length);
    Cor_CArg *carg = new Cor_CArg(tmp_buf);

    int vn = Cor_varname_enter(&(Cor_parserStruct->rule_var_names), carg);
    carg = new Cor_CArg();
    carg->set_var(vn);

    ca_list->push(carg->dup());

      // Just to be sure
    ASSERT(ca_list->count() == 2);
    pred = new Cor_Literal(Cor_parserStruct, *Cor_EquOpSymbol, 2);

    ca_list->build_arglist(pred->args);

    Cor_parserStruct->rhspreds.push((void *) pred);
    return(carg);
}
 
// *************************************************************************
Cor_CArg *Cor_create_multi_grouping(Cor_ParserStruct *parserStruct,
        Cor_CArgStack *ca_list)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_exEnv->dbg_aggregates) {
        *(Cor_exEnv->err_strm) << "in create relation count "
            << ca_list->count() << endl;
    }

    Cor_CArg *garg = new Cor_CArg();
    garg->set_grping(ca_list->count());

    garg->set_grp_subsum(0);
    garg->set_grp_seqnum(parserStruct->nested_index++);
    ca_list->build_arglist(garg->nested_args()->args());

    return(garg);
}

// *************************************************************************
void Cor_AddExportInfo(Cor_ParserStruct *Cor_parserStruct, Cor_CArg *arg)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (! arg->is_functor()) {
        *(Cor_exEnv->err_strm) << "export must be followed by a functor.";
        return;
    }

    for (int i = 0;; ) {     // Do at least once

           // LEAK
        Cor_ExportInfo *new_info = new Cor_ExportInfo();
        new_info->next = Cor_parserStruct->CurModule->ExportedPreds;
        new_info->pred_name = Cor_CArg(arg->functor());
        new_info->adorn_symbol.init();         // Arity == 0.
        new_info->sequence = Cor_parserStruct->CurModule->ExportedCount++;

        if (arg->arity() == 0)
           ;
        else if (! arg->args()[i].is_symbol()) {
            *(Cor_exEnv->err_strm) << "Bad kind of adornment for export "
                    << arg->functor() << ".\n";
        } else {

            new_info->adorn_symbol = arg->args()[i];
            delete new_info->adornment;
            new_info->adornment = Cor_create_adornment(new_info->adorn_symbol);
        }

        new_info->predicate = NULL;
        Cor_parserStruct->CurModule->ExportedPreds = new_info;

        if (++i >= arg->arity())
            break;
    }
}

// *************************************************************************
void COR_set_preprocessing(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.PreProcessing = val;
}

// *************************************************************************
void COR_set_no_magic(Cor_ModuleInfo& Module)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.UseSupplementaryMagic = 0;
    Module.UseMagic = 0;
}

// *************************************************************************
void COR_set_use_factormagic(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.UseFactorMagic = val;
}

// *************************************************************************
void COR_set_use_factoring(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.UseFactoring = val;
}

// *************************************************************************
void COR_set_supplementary_magic(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.UseSupplementaryMagic = val;
    if (val)
        Module.UseMagic = 0;
}

// *************************************************************************
void COR_set_ordinary_magic(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.UseMagic = val;
    if (val)
        Module.UseSupplementaryMagic = 0;
}

// *************************************************************************
void COR_set_rewriting(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.Rewriting = val;
}

// *************************************************************************
void COR_set_use_ordsearch(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.UseOrdSearch = val;
}

// *************************************************************************
void COR_set_lazy_eval(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.LazyEval = val;
}

// *************************************************************************
void COR_set_monotonic(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.Monotonic = val;
}

// *************************************************************************
void COR_set_pipelining(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.UsePipelining = val;
    if (val)
        COR_set_rewriting(Module, 0);
}

// *************************************************************************
void COR_set_check_subsum(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.CheckSubsum = val;
}

// *************************************************************************
void COR_set_multisets(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.MultiSet = val;
    if (val)
        Module.CheckSubsum = 0;
}

// *************************************************************************
void COR_set_disk_resident(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.DiskRel = val;
}

// *************************************************************************
void COR_set_index_deltas(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.IndexDeltas = val;
}

// *************************************************************************
void COR_set_single_scc(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.SingleScc = val;
}

// *************************************************************************
void COR_set_single_answer(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.SingleAnswer = val;
}

// *************************************************************************
void COR_set_convert_functions(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.ConvertFunctions = val;
}

// *************************************************************************
void COR_set_nested_scc_eval(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.NestedSccEval = val;
}

// *************************************************************************
void COR_set_naive_backtracking(Cor_ModuleInfo& Module, int val)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Module.NaiveBacktrack = val;
}

// *************************************************************************
//  Called from the parser : takes the information collected in the 
//  parserStruct and fills the cur_rule structure with it.

void Cor_complete_rule(Cor_Clause *cur_rule, Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete cur_rule->var_names;

    if (Cor_parserStruct->rule_var_names.count() == 0) {
        cur_rule->var_names = NULL;
    } else {
        cur_rule->var_names = Cor_parserStruct->rule_var_names.dup();
    }

    Cor_alloc_preds(cur_rule, Cor_parserStruct->rhspreds.length);
    
    simple_reorder_arith_preds(Cor_parserStruct);
    
      // Copy literals from parserStruct.rhspreds to new cur_rule
    for (int i= 0; i < Cor_parserStruct->rhspreds.length; ++i)
        cur_rule->preds[i] = (Cor_Literal*)Cor_parserStruct->rhspreds.data[i];

    Cor_parserStruct->cur_collect->push(cur_rule); 

    Cor_AddClauseToPredicate(cur_rule->head->predicate,cur_rule);
}

// *************************************************************************
static int check_builtin_op(Cor_CArg& name)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
     // This code assumes the order of the grouping part of the 
     // enum CArgKind matches the array builtin_ops[] in Gram.y

    for (int i=0; Cor_builtin_ops[i] != NULL; i++)
        if (name.equals(*(Cor_builtin_ops[i])))
            return(i);

    return(-1);
}

// *************************************************************************
static int is_convertible_pred(Cor_CArg& pred_name) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (check_builtin_op(pred_name) >= Cor_Builtin_Convertible_Start) 
        return(1);
    return(0);
}

// *************************************************************************
