// *************************************************************************
//    Unparse.C
// *************************************************************************

#include <stdio.h>
#include <iostream.h>

#include "Unparse.h"

#include "Literal.h"
#include "BuiltinSymbols.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************

static void print_unique_prednames(Cor_Clause *r, ostream *out);
static void sprint_adornment(Cor_BitVector& adorn, char *str, int *pos);

// *************************************************************************

int Cor_print_new_line_after_rule=1;

// *************************************************************************
// print_adorn_for_pred is cleared when rules need to be displayed for tracing.
// At this stage, the predicate names already have the adornment.
// When the magic program calls the routine, the predicate names do not have 
// their adornments yet, and so the adornment needs to be printed out. 
// The magic program uses the default setting of print_adorn_for_pred = 1.

int Cor_print_adorn_for_pred=1;
int Cor_C_suppress_all_free=0;

// **************************************************************************
void Cor_print_rule(Cor_Clause *rule, ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Cor_print_rule is outdated, use the stream operator instead.
    *out << *rule;


#ifdef  NOT_USED
    int j;
    int cn, nclauses;
    Cor_ClauseNames *cname = rule->clause_names;

    nclauses = cname->count();

    for (cn=0; cn < nclauses; cn++ ) {
         
            // Need to add quotes around labels
        *out << "\"";
        *out << cname->ith_symbol(cn);
        *out << "\"";

        if (cname->ith_neg(cn))
            *out << "-";
        *out << ": ";
    }

    Cor_print_pred(rule->head, out);
    for (int i=0; i < rule->head_deletes_count; i++) {
        *out << ", ";
        Cor_print_pred(rule->head_deletes[i], out, /* head */ 1);
    }
    if (rule->num_preds > 0 || rule->sips != NULL) {
        *out << " :- \n\t\t";
        for (j = 0; j < rule->num_preds; ++j) {
            Cor_print_pred(rule->preds[j], out);
            if (j < rule->num_preds - 1)
                *out << ", \n\t\t";     // print a pred on a line.
        }
        if (j==0)
            *out << "true";             // print "true" if no rhs preds
        if (rule->sips) {
            *out << " \n \t \t alias ";
            Cor_print_sips(rule, rule->sips, out);
        }
    }
    *out << ".";

    // Default is to print the new line. This is used by the magic program.
    // While dumping trace code, we dont want new lines printed. So then
    // the variable is cleared.
    if (Cor_print_new_line_after_rule)
        *out << endl;

#endif    // NOT_USED
}

// *************************************************************************
void Cor_print_rule_with_order(Cor_Clause *rule,
        int *predOrders, int *markOffsets, int recursive, ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;

    Cor_print_pred(rule->head, out);
    for(int i=0; i < rule->head_deletes_count; i++) {
        *out << ", ";
        Cor_print_pred(rule->head_deletes[i], out, /* head */ 1);
    }
    if (rule->num_preds > 0 || rule->sips != NULL) {
        *out << " :- \n\t\t";
        for (j = 0; j < rule->num_preds; ++j) {
            if ((markOffsets[2*predOrders[j]+1] % 2) == 0) // print old 
                *out << "old_";
            if ((j==0) && recursive)
                *out << "d_";       // print delta
            Cor_print_pred(rule->preds[predOrders[j]], out);
            if (j < rule->num_preds - 1)
            *out << ", \n\t\t";     // print a pred on a line.
        }

        if (j==0)
            *out << "true";         // print "true" if no rhs preds

//    if (rule->sips) {
//        *out << " \n \t \t alias ";
//        print_sips(rule, rule->sips, out);
//    }
    }
    *out << ".";

// Default is to print the new line. This is used by the magic program.
// While dumping trace code, we dont want new lines printed. So then
// the variable is cleared.
    if (Cor_print_new_line_after_rule)
         *out << endl;
}

// *************************************************************************
static void sprint_pred( Cor_Literal *p, char *str, int *pos, int head=0)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (p->kind) {
    case COR_REGULAR:
        if ( p->negated ) {
            if (head) 
                sprintf(str +*pos," del ");
            else 
                sprintf(str +*pos," not ");
            *pos += 5;
        }
        break;
    default:
        ASSERT((! p->negated ));
        break;
    }

    sprintf(str+(*pos), "%s", p->pred.symbol());
    *pos += p->pred.symlen();

    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    sprint_adornment(p->adorn, str, pos);
    Cor_C_suppress_all_free = temp;
    sprintf(str + *pos, "(");
    (*pos)++;
    p->args.sprint(str, pos);
    sprintf(str + *pos, ")");
    (*pos)++;
}

// *************************************************************************
void Cor_sprint_rule(Cor_Clause *rule, char *str)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;
    int str_pos = 0;

    int cn, nclauses;
    Cor_ClauseNames *cname = rule->clause_names;

    nclauses = cname->count();

    for (cn=0; cn < nclauses; cn++ ) {
         
        sprintf(str+str_pos, "%s", cname->ith_symbol(cn));
        str_pos += strlen(cname->ith_symbol(cn));
        if (cname->ith_neg(cn)) {
            sprintf(str+str_pos, "-");
            str_pos++;
        }
        sprintf(str+str_pos, ": ");
        str_pos += 2;
    }

    sprint_pred(rule->head, str, &str_pos);
    for (j=0; j < rule->head_deletes_count; j++) {
        sprintf(str+str_pos, ", ");
        str_pos += 2;
        sprint_pred(rule->head_deletes[j], str, &str_pos, /* head */ 1);
    }
    if (rule->num_preds > 0 || rule->sips != NULL) {
        sprintf(str+str_pos, " :- ");
        str_pos += 4;
        for (j = 0; j < rule->num_preds; ++j) {
            sprint_pred(rule->preds[j], str, &str_pos);
            if (j < rule->num_preds - 1) {
                sprintf(str+str_pos, ", \n\t\t");
                str_pos += 5;
            }
        }
        if (j == 0){
            sprintf(str+str_pos, "true");
            str_pos += 4;
        }

        // NOTE : Left out sips information
    }

    sprintf(str+str_pos, ".");
}

// *************************************************************************
/*
    Print a unique (within 'rule') name for the index'th predicate of 'rule',
    where index==0 means the head, and 1 is the first RHS predicate.
*/

static void Cor_print_unique_predname(Cor_Clause *rule, int index, ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *pred = index ? rule->preds[index-1] : rule->head;

#    ifdef SHORT_SIP_PREDNAMES
        *out <<  (pred->pred.symbol())[0] <<  index;
#    else
        char num_preds[12];
        sprintf(num_preds, "%d", rule->num_preds);
        *out <<  pred->pred.symbol();
        // Letting the format depend on num_preds guarantees that the
        // same number of digits is used for all the preds.
        //*out << setw(strlen(num_preds)) << index << setw();
        *out << index;
#    endif
}

// *************************************************************************
static void print_unique_prednames(Cor_Clause *r, ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    *out << "[";
    Cor_print_unique_predname(r, 0, out);
    for (i = 1; i <= r->num_preds; i++) {
        *out << ", ";
        Cor_print_unique_predname(r, i, out);
    }
    *out << "]";
}

// *************************************************************************
void Cor_print_pred( Cor_Literal *p, ostream *out, int head)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    switch (p->kind) {
    case COR_REGULAR:
        if ( p->negated ) {
            if (head) 
                *out << " del ";
            else 
                *out << " not ";
        }
        p->pred.printon(out);
        break;
    case COR_MAGIC:
        // CHANGE
        ASSERT((! p->negated ));
        *out << "m_" <<  p->pred.symbol();
        break;
    case COR_SUPMAGIC:
        // CHANGE
        ASSERT((! p->negated ));
        *out <<  p->pred.symbol();
        break;
    case COR_FACTORMAGIC:
        ASSERT((! p->negated ));
        *out << "mf_" <<  p->pred.symbol();
        break;
    case COR_DONE_MAGIC:
        ASSERT((! p->negated ));
        *out << "done_m_" <<  p->pred.symbol();
        break;
    }

    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    Cor_print_adornment(p->adorn, out);
    Cor_C_suppress_all_free = temp;
    *out << "(";
    p->args.printon(out);
    *out << ")";
}

// *************************************************************************
void Cor_print_adornment_brack(Cor_BitVector& a, ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;

    for (i = 0; i < a.len(); ++i) {
        *file << (a.test(i) ? "b" : "f");
    }
}

// *************************************************************************
void Cor_print_sip(Cor_Clause *r, Cor_Sip *s, ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int b;

    for (; s; s = s->next) {
        *out << "\n\t\t\t\t[";
        for (b = s->tail.next(-1); b >= 0; b = s->tail.next(b)) {
            Cor_print_unique_predname(r, b, out);
            if (s->tail.next(b) >= 0 ) 
                *out << ", ";
        }
        *out << "] ==> ";
        Cor_print_unique_predname(r, s->head, out);
        *out << "(";
        for (int c= s->bound_vars.next(-1),i=0; c>=0;
            c=s->bound_vars.next(c), i++)
        {
            *out << *(r->var_names->get_ith(c));
            if ( i )
                *out << ",";
        }
        *out << ")";

        if (s->next)
            *out << ",\n";
        else
            *out << endl;
    }
}

// *************************************************************************
void Cor_print_sips(Cor_Clause *r, Cor_Sips *s, ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int sip_printed = 0;
    print_unique_prednames(r, out);
    *out << endl;
    if ( s )
        *out << "\t\t sips [\n ";

    for (; s; s = s->next) {
        if( 1 /*s->sip_list*/ ) {
            if (sip_printed)
                *out << ",\n";
            sip_printed = 1;
            *out << "\t\t\t[ ";
            Cor_print_adornment_brack(s->adorn, out);
            *out << ": ";
            Cor_print_sip(r, s->sip_list, out);
            *out << "\t\t\t]";
        }
    }

    if ( !sip_printed )
        *out << "\t[ ]";
    else
        *out << "\t\t ]\n";
}

// *************************************************************************
static int all_args_free(Cor_BitVector *a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return(allargsfree(a));
}

// *************************************************************************
static int all_args_free(char *a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (; *a; a++)
        if (*a == 'b')
            return(0);
    return(1);
}

// *************************************************************************
static int suppress_printing(Cor_BitVector *a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (! Cor_print_adorn_for_pred)
        return 1;

    if (a->len() == 0)
        return 1;

    if (Cor_C_suppress_all_free && all_args_free(a))
        return 1;

    return 0;
}

// *************************************************************************
static int suppress_printing(char *a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (! Cor_print_adorn_for_pred)
        return 1;
    if (strlen(a) == 0)
        return 1;
    if (Cor_C_suppress_all_free && all_args_free(a))
        return 1;
    return 0;
}

// *************************************************************************
void sprint_adornment(char *buf, Cor_BitVector *a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!a) return;

    if (! suppress_printing(a)) {
        *buf++ = '_';
        for (int i = 0; i < a->len(); i++)
            buf[i] = a->test(i) ? 'b' : 'f';
        buf[a->len()] = '\0';
    }
}

// *************************************************************************
void sprint_adornment(char *buf, char *a)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if(!a) return;

    if ( ! suppress_printing(a)) {
        *buf++ = '_';
        sprintf(buf, "%s", a);
    }
}

// *************************************************************************
static void sprint_adornment(Cor_BitVector& a, char *str, int *pos)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ( ! suppress_printing(&a)) {
        sprintf(str + *pos, "_"); (*pos)++;
        for (int i = 0; i < a.len(); ++i) {
            sprintf(str + *pos, a.test(i) ? "b" : "f");
            (*pos)++;
        }
    }
}

// *************************************************************************
void Cor_print_adornment(Cor_BitVector& a, ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!suppress_printing(&a)) {
        *file << "_";
        for (int i = 0; i < a.len(); ++i)
            *file << (a.test(i) ? "b" : "f");
    }
}

// *************************************************************************
void Cor_print_adornment(Cor_BitVector *a, ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!a) return;
    Cor_print_adornment(*a, file);
}

// *************************************************************************
void Cor_print_adornment(char *a, ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if(!a) return;
    if ( !suppress_printing(a)) {
        *file << "_" << a;
    }
}

// *************************************************************************
void Cor_sprint_magic_name(char *buf, char *name, Cor_BitVector *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    sprintf(buf, "m_%s", name);
    char *p = buf + strlen(buf);
    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    sprint_adornment(p,adorn);
    Cor_C_suppress_all_free = temp;
}

// *************************************************************************
void Cor_sprint_magic_name(char *buf, char *name, char *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    sprintf(buf, "m_%s", name);
    char *p = buf + strlen(buf);
    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    sprint_adornment(p,adorn);
    Cor_C_suppress_all_free = temp;
}

// *************************************************************************
void Cor_fprint_magic_name(ostream *out, char *name, Cor_BitVector *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << "m_" << name;
    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    Cor_print_adornment(adorn, out);
    Cor_C_suppress_all_free = temp;
}

// *************************************************************************
void Cor_fprint_magic_name(ostream *out, char *name, char *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << "m_" << name;
    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    Cor_print_adornment(adorn, out);
    Cor_C_suppress_all_free = temp;
}

// *************************************************************************
void Cor_sprint_pred_name(char *buf, char *name, Cor_BitVector *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    sprintf(buf, "%s", name);
    char *p = buf + strlen(buf);
    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    sprint_adornment(p,adorn);
    Cor_C_suppress_all_free = temp;
}

// *************************************************************************
void Cor_sprint_pred_name(char *buf, char *name, char *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    sprintf(buf, "%s", name);
    char *p = buf + strlen(buf);
    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    sprint_adornment(p,adorn);
    Cor_C_suppress_all_free = temp;
}

// *************************************************************************
void Cor_fprint_pred_name(ostream *out, char *name, Cor_BitVector *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << name;
    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    Cor_print_adornment(adorn, out);
    Cor_C_suppress_all_free = temp;
}

// *************************************************************************
void Cor_fprint_pred_name(ostream *out, char *name, char *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << name;
    int temp = Cor_C_suppress_all_free;
    Cor_C_suppress_all_free = 1;
    Cor_print_adornment(adorn, out);
    Cor_C_suppress_all_free = temp;
}

// *************************************************************************
void Cor_print_allowed_adornment_annotations(Cor_PredAnnotations *anno,
        ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_PredAnnotations *cur;

    for (cur = anno; cur; cur=cur->next()) {
        *file << "@" << Cor_AllowedAdornSymbol->symbol() << "  ";
        cur->print(file, 0); // adornment printing wrt non-adorned predicates
        *file << ".\n";
    }
}

// *************************************************************************
void Cor_print_aggsel_annotations(Cor_PredAnnotations *agg_sel, ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (Cor_PredAnnotations *cur = agg_sel; cur; cur=cur->next()) {
        *file << "@";
        switch (cur->agg_type) {
        default:
        case AggSel_Regular:
            *file << *Cor_AggSelectionSymbol;
             break;
        case AggSel_PerIter:
            *file << *Cor_AggSel_PerIter_Symbol;
             break;
        case AggSel_MultAnswers:
            *file << *Cor_AggSel_MultAnswers_Symbol;
             break;
        }
        *file << "  ";
        cur->print(file);
        *file << ".\n";
    }
}

// *************************************************************************
void Cor_print_general_annotations(Cor_PredAnnotations *anno,
        Cor_CArg *AnnoSym, ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (Cor_PredAnnotations *cur = anno; cur; cur=cur->next()) {
        *file << "@" << *AnnoSym << "  ";
        cur->print(file);
        *file << ".\n";
    }
}

// *************************************************************************
void Cor_print_MultisetAnnotations(Cor_PredAnnotations *anno, ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_PredAnnotations *cur;
    
    for (cur = anno; cur; cur=cur->next()) {
        if (cur->check_subsum == 0) 
            *file << "@" << Cor_MultisetSymbol->symbol() << "    ";
        else
            *file << "@" << Cor_CheckSubsumSymbol->symbol() << "    ";
        
        cur->print(file, 0);
        *file << ".\n";
    }

}

// *************************************************************************
void Cor_print_Annotations(Cor_ModuleInfo& module_info, ostream *file,
    Cor_PredAnnotations *index, Cor_PredAnnotations *agg_sel,
    Cor_PredAnnotations *allowed_adorn, Cor_PredAnnotations *multiset,
    Cor_PredAnnotations *diskrel, Cor_PredAnnotations *asyncrel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *file << "@" << Cor_NoPreProcessingSymbol->symbol()
          << (module_info.PreProcessing ? '-' : '+') << ".\n";
    *file << "@" << Cor_SupplementaryMagicSymbol->symbol()
          << (module_info.UseSupplementaryMagic ? '+' : '-') << ".\n";
    *file << "@" << Cor_UseFactoringSymbol->symbol()
          << (module_info.UseFactoring ? '+' : '-') << ".\n";
    *file << "@" << Cor_OrdinaryMagicSymbol->symbol()
          << (module_info.UseMagic ? '+' : '-') << ".\n";
    *file << "@" << Cor_NoRewritingSymbol->symbol()
          << (module_info.Rewriting ? '-' : '+') << ".\n";
    *file << "@" << Cor_UseOrdSearchSymbol->symbol()
          << (module_info.UseOrdSearch ? '+' : '-') << ".\n";
    *file << "@" << Cor_UsePipeliningSymbol->symbol()
          << (module_info.UsePipelining ? '+' : '-') << ".\n";
    *file << "@" << Cor_SingleSccSymbol->symbol()
          << (module_info.SingleScc ? '+' : '-') << ".\n";
    *file << "@" << Cor_SingleAnswerSymbol->symbol()
          << (module_info.SingleAnswer ? '+' : '-') << ".\n";
    *file << "@" << Cor_NaiveBacktrackSymbol->symbol()
          << (module_info.NaiveBacktrack ? '+' : '-') << ".\n";
    *file << "@" << Cor_NestedSccEvalSymbol->symbol()
          << (module_info.NestedSccEval ? '+' : '-') << ".\n";

    Cor_print_general_annotations(index, Cor_MakeIndexSymbol, file);
    Cor_print_aggsel_annotations(agg_sel, file);
    Cor_print_allowed_adornment_annotations(allowed_adorn, file);
    Cor_print_general_annotations(diskrel, Cor_DiskResidentSymbol, file);
    Cor_print_general_annotations(asyncrel, Cor_ASyncRelSymbol, file);

    *file << "@" << *Cor_CheckSubsumSymbol
          << (module_info.CheckSubsum ? '+' : '-') << ".\n";
    *file << "@" << *Cor_MultisetSymbol
          << (module_info.MultiSet ? '+' : '-') << ".\n";
    Cor_print_MultisetAnnotations(multiset, file);


    *file << "@" << *Cor_IndexDeltaSymbol
          << (module_info.IndexDeltas ? '+' : '-') << ".\n";
    *file << "@" << *Cor_ConvertFunctionsSymbol
          << (module_info.ConvertFunctions ? '+' : '-') << ".\n";
    *file << "@" << *Cor_LazyEvalSymbol
          << (module_info.LazyEval ? '+' : '-') << ".\n";
    *file << "@" << *Cor_MonotonicSymbol
          << (module_info.Monotonic ? '+' : '-') << ".\n";
    *file << "@" << *Cor_DiskResidentSymbol
          << (module_info.DiskRel ? '+' : '-') << ".\n";
}

// *************************************************************************
