#ifdef SOCK_REL

// *************************************************************************
//    SockRelation.C
// *************************************************************************

#include "SockRelation.h"

#include "Util.h"
#include "Globals.h"
#include "CoralError.h"
#include "Unify.h"
#include "Debug.h"
#include "TupleIterator.h"
#include "ClientProtocol.h"

// *************************************************************************
Cor_SockRelation::Cor_SockRelation(char *ln, int arity, char * host, int port, char * ws, char * rn)
    : Cor_StorageRelation(arity) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    strcpy(_local_name,ln);
    _arity = arity;
    strcpy(_host,host);
    _port = port;
    strcpy(_work_space,ws);
    strcpy(_relation_name, rn);

    // turn off the checking for subsumption. it's done on remote server
    check_subsum = 0;

    Cor_sockbuf * sbuf = new Cor_sockbuf(_host);
    valid = sbuf->valid();
    if (!valid) {
        *(Cor_exEnv->err_strm)
            << "Error: could not get connected with coral server "
            << _host << endl;
        cs = NULL; 
        return;
    }
    else {
        *(Cor_exEnv->err_strm) << "Connected to coral server " << _host << endl;
        cs = new CoralStream(sbuf);
        sync_cnt = 2;
        
        // switch to the correct work space
        *cs << "clear(interactive_mode).\n";
        *cs << "clear(verbose_answer).\n";
        *cs << "assign(answer_style,3).\n";
        *cs << "\nsync(" << sync_cnt << ").\n";
        cs->flush();

        ReceiveResponse();

        *cs << "ws_create(\"" << _work_space << "\").";
        *cs << "ws_current(\"" << _work_space << "\").";
        *cs << "\nsync(" << sync_cnt << ").\n";
        cs->flush();

        ReceiveResponse();
    }
}

// *************************************************************************
Cor_SockRelation::~Cor_SockRelation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (cs) delete cs;
}

// *************************************************************************
void Cor_SockRelation::ReceiveResponse()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DEBUG_SOCK_REL
    int quiet = 0;
#else
    int quiet = 1;
#endif
    
    int done = 0;
    int sync = 0;
    int ok_cnt = 2;

    int have_sent_yield = 0;

    do {
        memset(Buf,'\0', 1000);

        if (!(cs->gets(Buf, 1000)))
            break;

#ifdef DEBUG_SOCK_REL
        *(Cor_exEnv->out_strm) << "Buf = " << Buf;
#endif
        
        if (!have_sent_yield) {
            *cs << "yield.\n";
            cs->flush();
            have_sent_yield = 1;
        }

        if (Buf[4] == ':'){
            if (strncmp(Buf,CP_OKAY,5) == 0) {
                if (sync)
                    if (--ok_cnt == 0)
                        done = 1;
            }
            else if (strncmp(Buf,CP_SYNC,5) == 0) {
                int val;
                sscanf(&(Buf[5]), "%d", &val);
                if (val == sync_cnt){
                    sync = 1;
                    sync_cnt++;
                }
            }
            else if (strncmp(Buf,CP_CERR,5) == 0) {
                if (!quiet) 
                    *(Cor_exEnv->out_strm) << (&(Buf[5]));

            }
            else if (strncmp(Buf,CP_MESG,5) == 0) {
                *(Cor_exEnv->out_strm) << (&(Buf[5]));

            } else if (strncmp(Buf,CP_LOGIN,5) == 0) {
            } else if (strncmp(Buf,CP_USER,5) == 0) {
                *(Cor_exEnv->out_strm) << (Buf);

            } else if (strncmp(Buf,CP_PSWD,5) == 0) {
                *(Cor_exEnv->out_strm) << (Buf);
                // Save the password for use with Explain
                strncpy(PassWord,&(Buf[5]),8);
                PassWord[8] = '\0';
                int i = strlen(PassWord);
                if (PassWord[i-1] == '\n')
                    PassWord[i-1] = '\0';

            } else if (strncmp(Buf,CP_R_ID,5) == 0) {
                *(Cor_exEnv->out_strm) << (Buf);

            } else if (strncmp(Buf,CP_PRMT,5) == 0) {

                strcpy(Prompt,&(Buf[5]));
                Prompt[strlen(Prompt)-1] = ' ';

            } else if (strncmp(Buf,CP_EXIT,5) == 0) {

                *(Cor_exEnv->out_strm) << ("exiting.\n");
                return;

            } else {
                if (!quiet) { 
                    *(Cor_exEnv->err_strm) << ("Warning, unimplemented protocol header.\n");
                    *(Cor_exEnv->out_strm) << (Buf);
                }
            }
        } else {
            if (!quiet) 
                *(Cor_exEnv->out_strm) << (Buf);
        }

    } while(!done);

}

// ************************************************************************
void Cor_SockRelation::Init(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.ipos == Scanning_Init) {
        // send the query to the server
        *cs << "?" << _relation_name << "(";
        for (int i = 0; i < iterator.arg_list.arity(); i++)  {
            Cor_Term term(iterator.arg_list[i], iterator.bindenv);
            Cor_deep_dereference_term(term);
            *cs << term.expr;
            if (i == iterator.arg_list.arity() - 1)
                *cs << ')';
            else
                *cs << ',';
        }
        *cs << '.';
        *cs << "\nsync(" << sync_cnt << ").\n";
        cs->flush();
        
        iterator.stack_mark.get_mark();
        iterator.ipos = Scanning_Known;

    }
    else {
        iterator.stack_mark.pop_to();

    }

}

// ************************************************************************
Cor_Tuple * Cor_SockRelation::Done(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    iterator.ipos = Scanning_Done;
    iterator.set_no_match();

    return NULL;
}

// ************************************************************************
Cor_Tuple *Cor_SockRelation::GetNext()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DEBUG_SOCK_REL
    int quiet = 0;
#else
    int quiet = 1;
#endif
    
    int done = 0;
    int sync = 0;
    int ok_cnt = 2;

    int have_sent_yield = 0;

    do {
        memset(Buf, '\0', 1000);

        if (!(cs->gets(Buf, 1000)))
            break;

#ifdef DEBUG_SOCK_REL
        *(Cor_exEnv->out_strm) << "Buf = " << Buf;
#endif

        if (!have_sent_yield) {
            *cs << "yield.\n";
            cs->flush();
            have_sent_yield = 1;
        }

        if (Buf[4] == ':'){
            if (strncmp(Buf,CP_OKAY,5) == 0) {
                if (sync)
                    if (--ok_cnt == 0)
                        done = 1;
            }
            else if (strncmp(Buf,CP_SYNC,5) == 0) {
                int val;
                sscanf(&(Buf[5]), "%d", &val);
                if (val == sync_cnt){
                    sync = 1;
                    sync_cnt++;
                }
            }
            else if (strncmp(Buf,CP_CERR,5) == 0) {
                if (!quiet) 
                    *(Cor_exEnv->out_strm) << (&(Buf[5]));

            }
            else if (strncmp(Buf,CP_MESG,5) == 0) {
                *(Cor_exEnv->out_strm) << (&(Buf[5]));

            } else if (strncmp(Buf,CP_LOGIN,5) == 0) {
            } else if (strncmp(Buf,CP_USER,5) == 0) {
                *(Cor_exEnv->out_strm) << (Buf);

            } else if (strncmp(Buf,CP_PSWD,5) == 0) {
                *(Cor_exEnv->out_strm) << (Buf);
                //Save the password for use with Explain
                strncpy(PassWord,&(Buf[5]),8);
                PassWord[8] = '\0';
                int i = strlen(PassWord);
                if (PassWord[i-1] == '\n')
                    PassWord[i-1] = '\0';

            } else if (strncmp(Buf,CP_R_ID,5) == 0) {
                *(Cor_exEnv->out_strm) << (Buf);

            } else if (strncmp(Buf,CP_PRMT,5) == 0) {

                strcpy(Prompt,&(Buf[5]));
                Prompt[strlen(Prompt)-1] = ' ';

            } else if (strncmp(Buf,CP_EXIT,5) == 0) {

                *(Cor_exEnv->out_strm) << ("exiting.\n");
                return NULL;

            } else {
                if (!quiet) {
                    *(Cor_exEnv->err_strm) << ("Warning, unimplemented protocol header.\n");
                    *(Cor_exEnv->out_strm) << (Buf);
                }
            }
        } else {
            if (!quiet) 
                *(Cor_exEnv->out_strm) << (Buf);
            if (!strncmp(_relation_name, Buf, strlen(_relation_name))) {
                Cor_CArg relname(_relation_name);
                int ar = _arity;

                Cor_Tuple * tuple = Cor_parseFact(Buf,relname, &ar);
                if (tuple) {
#ifdef DEBUG_SOCK_REL
                    *(Cor_exEnv->out_strm) << "TUPLE = " << *tuple << endl;
#endif
                    return tuple;
                }
            }
        }

    } while(!done);

    return NULL;
}
    

// ************************************************************************
Cor_Tuple *Cor_SockRelation::get_next_tuple(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DEBUG_SOCK_REL
    *(Cor_exEnv->out_strm) << "Getting the  next tuple from relation " << this << "\n";
#endif
    
    if (Cor_exEnv->C_interrupt_raised)
      return Done(iterator);

    Init(iterator);

    Cor_Tuple *tuple = GetNext();

    if (tuple && Cor_unify_literal(iterator, tuple, Cor_exEnv->dbg_get_next)) {
        iterator.reset_no_match();
        return tuple;
    }
    else 
        return Done(iterator);
    
}

// *************************************************************************
Cor_BindEnv *Cor_SockRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Tuple *result = get_next_tuple(iterator);
    if (!result) return NULL;
    return iterator.bindenv;
}

// *************************************************************************
int Cor_SockRelation::insert_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DEBUG_SOCK_REL
    *(Cor_exEnv->out_strm) << "Inserting tuple " << *tuple << " into relation " << this << endl;
#endif
    
    *cs << "insert(" << _relation_name << "(" << *tuple << ")).";
    *cs << "\nsync(" << sync_cnt << ").\n";
    cs->flush();

    ReceiveResponse();

    return 0; // let the caller decide whether to delete the tuple
}

// *************************************************************************
void Cor_SockRelation::print_name(ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << "local_name = " << _local_name << endl;
    *out << "arity = " << _arity << endl;
    *out << "host = " << _host << endl;
    *out << "port = " << _port << endl;
    *out << "work_space = " << _work_space << endl;
    *out << "relation_name = " << _relation_name << endl;
}

// *************************************************************************
void Cor_SockRelation::release(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Done(iterator);
}

// *************************************************************************
void Cor_SockRelation::empty_relation(int deleteTuples)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DEBUG_SOCK_REL
    *(Cor_exEnv->out_strm) << "Emptying the relation " << _local_name << endl;
#endif

    if (!deleteTuples) return;
    
    // ??? it's really a relation on the remote server. we should not empty it.
//     *cs << "rel_clear(" << _relation_name << '(';
//     for (int i = 0; i < _arity - 1; i++) 
//         *cs << "_,";
//     *cs << "_)).";
//     *cs << "\nsync(" << sync_cnt << ").\n";
//     cs->flush();
    
//     ReceiveResponse();
}

// *************************************************************************
int Cor_SockRelation::tuple_delete(Cor_Tuple* tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DEBUG_SOCK_REL
    *(Cor_exEnv->out_strm) << "Deleting tuple " << *tuple << " from relation " << this << endl;
#endif

    *cs << "delete(" << _relation_name << "(" << *tuple << ")).";
    *cs << "\nsync(" << sync_cnt << ").\n";
    cs->flush();

    ReceiveResponse();

    return 1; 
}

// *************************************************************************

#endif // SOCK_REL
