// *************************************************************************
//    Predicate.h
// *************************************************************************

#ifndef _Predicate_H_
#define _Predicate_H_

#include "Scc.h"
#include "SccWork.h"
#include "Table.h"
#include "Compile.h"


// *************************************************************************

void Cor_AddClauseToPredicate(Cor_Predicate *predicate, Cor_Clause *clause);

int Cor_is_magic(Cor_CArg& name);
int Cor_is_mp_done(Cor_CArg& name);
int Cor_is_supp(Cor_CArg& name);

// *************************************************************************
// A Predicate is identified by its name and arity. The other fields
// are used by various pieces of code that manipulate Predicates.

class Cor_Predicate {

  public:

    Cor_CArg       name;
    int            arity;
    Cor_Predicate *next;         // link to next predicate in the same Scc 
    Cor_Predicate *symtabnext;   // link to next predicate of same name 
                                 // in the symbol table
    Cor_Predicate *allprednext;  // link to next predicate in AllPredicates list
    Cor_Clause    *clauses;
    Cor_Clause   **clause_tail;
    Cor_Scc       *scc;

    Cor_SccWork        *work;     // temporaries used for Scc generation 
    char                mark;     //  ''

    int         frame_offset;     // Set and used during evaluation. 
    int         ext_magic;        // true if clauses point to an ext magic rule

#ifdef ASYN_SOCK_REL
    int asyn_sock();              // if the predicate represents an
                                  // asyn_sock_rel

    int delta_done();             // the value of _delta_done for the
                                  // associated relation
#endif
    
    inline int Cor_is_magic()    { return ::Cor_is_magic(name); }
    inline int Cor_is_mp_done()  { return ::Cor_is_mp_done(name); }
    inline int Cor_is_supp()     { return ::Cor_is_supp(name); }

    Cor_Predicate();
    Cor_Predicate(Cor_ParserStruct *ps, Cor_CArg& name, int arity);

   ~Cor_Predicate();
};

// *************************************************************************

#ifndef OFFSET_HACK
inline int Cor_get_offset(Cor_Table *F, Cor_Predicate *pred,
        int print_error = 1)
    {
        return Cor_get_offset(F, pred->name, pred->arity, print_error) ;
    }
#else
    // WARNING: This code should only be called after predicates have been
    // inserted into the frame table.

inline int Cor_get_offset(Cor_Table *F, Cor_Predicate *pred,
        int print_error = 1)
    {
        ASSERT((pred->frame_offset >= 0) && (pred->frame_offset < F->numRels));
        return pred->frame_offset;
    }
#endif

// *************************************************************************

#endif _Predicate_H_
