// *************************************************************************
//    RuleEnv.C
// *************************************************************************

#include "RuleEnv.h"

#include "TupleIterator.h"
#include "Globals.h"
#include "Debug.h"

// *************************************************************************
Cor_RuleEnv::Cor_RuleEnv(Cor_SemiNaive *parent)
         : bindenv(parent->rInfo->num_vars) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    max_literals = parent->rInfo->num_literals;

    // This is a hack to prevent memory overwrites for rules of the form p(X).
    // rInfo->num_literals == 0 for these rules, there are places in the code
    // where it is assumed that at least one rhs literal exists.  This should
    // actually be fixed by either changing that code, or by introducing a
    // true literal on the rhs.

    if (!max_literals) max_literals = 1;

    rmarks = new Cor_StackMark [max_literals];

    tuple_its = new Cor_TupleIterator * [max_literals];

    arglist_its = new Cor_ArgList * [max_literals];
    tuples = new Cor_Tuple * [max_literals];
    bindenvs = new Cor_BindEnv *[max_literals];
    parent_sn = parent;

    // allocate tuple iterators if possible
    if (parent->rule_id >= 0) {
        for (int i = 0; i < max_literals; i++) {
            tuple_its[i] =
                new Cor_TupleIterator(NULL, parent->rInfo->arg_list[i],
                NULL, parent->unify_funct[i], NULL, NULL);
        }
    } else {
        // Need to set them to something so they can be safely deleted
        for (int i = 0; i < max_literals; i++)
            tuple_its[i] = NULL;
    }
}

// *************************************************************************
Cor_RuleEnv::~Cor_RuleEnv()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete [] rmarks;

      // delete each of the tuple iterators !
    for (int i = 0; i < max_literals; i++)
      delete tuple_its[i];

    delete [] tuple_its;
    delete [] arglist_its;
    delete [] tuples;
    delete [] bindenvs;
}

// *************************************************************************
// Clear out the tuple iterators
void Cor_RuleEnv::clear_tupits()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // delete each of the tuple iterators !
    for (int i = 0; i < max_literals; i++) {
        delete tuple_its[i];
        tuple_its[i] = NULL;
    }
}

// *************************************************************************
