// **************************************************************************
//    hsr.C
// **************************************************************************

#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <iostream.h>

#include "CoralInit.h"
#include "Globals.h"
#include "ParseStack.h"
#include "CoralError.h"
#include "Util.h"
#include "BuiltinUtil.h"

#include "LocalConnectArray.h"
#include "Debug.h"

#include "Interface.h"

// *************************************************************************
// *************************************************************************

// Special includes needed just for the tests
#include "HashSimpleRelation.h"
#include "TupleIterator.h"


// *************************************************************************
 // You can do whatever you like inside here, it is called once.

void run_test()
{
    int        j, rair;
    Cor_CArg   rel_nme;
    Cor_Tuple *tuple;
    Cor_HashSimpleRelation *hsr = new Cor_HashSimpleRelation(3);
    Cor_RMark *rmark=NULL, *startmark, *endmark; 
    Cor_RMark *rmark2;

    char *data[] = {
            "veg(1,5,2)",
            "veg(2,beans,rice)",
            "veg(3,m,[])",
            "veg(4,wheat,potato)",
            "veg(5,beans,lemons)",
            NULL
        };

    char *more_data[] = {
            "veg(6,6,2)",
            "veg(7,3,rice)",
            "veg(8,2,[])",
            "veg(9,wheat,10)",
            "veg(10,beans,112)",
            NULL
        };

    hsr->delta_indexed = 0;


    startmark = hsr->getMark();

    if (hsr->delta_indexed)
        cout << "Relation is delta_indexed.\n";
    else
        cout << "Relation is not delta_indexed.\n";

    for (j=0; data[j]; j++) {

        tuple = Cor_parseFact(data[j], rel_nme, &rair); 
        if (! hsr->insert_tuple(tuple))
            cout << "Insert of " << *tuple << " failed.\n";
        else
            cout << "Inserted tuple: " << *tuple << endl;

        if (j==2) {
            rmark = hsr->getMark();
        }
    }

    rmark2 = hsr->getMark();

    hsr->freeMark(startmark);
    startmark = rmark;

    for (j=0; more_data[j]; j++) {
        tuple = Cor_parseFact(more_data[j], rel_nme, &rair); 
        if (! hsr->insert_tuple(tuple))
            cout << "Insert of " << *tuple << " failed.\n";
        else
            cout << "Inserted tuple: " << *tuple << endl;
    }

    endmark = hsr->getMark();

    cout << "\nInternals:\n";
    hsr->show_internals(cout);

    cout << "\nPartial Scan:\n";
    Cor_Tuple *qtuple = Cor_parseFact("q(X,Y,Z)", rel_nme, &rair); 

    Cor_TupleIterator *it = new Cor_TupleIterator(hsr,qtuple,
            NULL, startmark, rmark);

    it->bindenv = new Cor_BindEnv(3);
    it->reset();

    for ( tuple = it->get_next_tuple();
                    !it->no_match(); tuple = it->get_next_tuple()) {
          // Print them out as tuples
        if (tuple)
            cout << "veg(" << *tuple << ").\n";
        else
            cout << "NULL returned.\n";
    }

    delete it;

    cout << "\nRest Partial Scan:\n";
    it = new Cor_TupleIterator(hsr,qtuple, NULL, rmark, endmark);
    it->bindenv = new Cor_BindEnv(3);
    it->reset();

    for ( tuple = it->get_next_tuple();
                    !it->no_match(); tuple = it->get_next_tuple()) {
          // Print them out as tuples
        if (tuple)
            cout << "veg(" << *tuple << ").\n";
        else
            cout << "NULL returned.\n";
    }

    hsr->freeMark(startmark);
    hsr->freeMark(endmark);

    cout << "\nTuples:\n";
    Cor_ScanDesc *scan = new Cor_ScanDesc(hsr);

      // Iterate over the tuples in the relation
    for (tuple = scan->next_tuple(); !(scan->no_match());
                     tuple = scan->next_tuple()) {

          // Print them out as tuples
        cout << "veg(" << *tuple << ").\n";
    }
    delete scan;

    delete hsr;
}

// *************************************************************************
// *************************************************************************
// You should never need change main()
int main(int ac, char **av)
{
      // We need these right away
    Cor_new_globals();

      // Start up a "Local" Connection (use cin_strm/cout_strm)
    Cor_clients = new Cor_LocalConnectArray();

    Cor_exEnv->C_interactive_mode_default = 0;
    Cor_exEnv->C_exec_mode = COR_DOING_INTERPRET;

    init_coral(av[0]);
    init_coral_interactive();

    // *********************************

    cout << "\nCalling run_test()\n";

    run_test();

    cout << "Done calling run_test()\n";

    // *********************************
    
    if (!Cor_exEnv->C_quiet_mode_default) {
        cerr << "\nHave a good day, " << cuserid(NULL) << "!\n";
    }

    exit_coral();
    return(0);
}

//**************************************************************************
