// *************************************************************************
//    Rewriting.C
// *************************************************************************

#include <stdio.h>
#include <unistd.h>

#include "Rewriting.h"

#include "Globals.h"
#include "CoralError.h"
#include "ParserStruct.h"
#include "ExportInfo.h"
#include "Unparse.h"
#include "Util.h"
#include "Collection.h"
#include "BuiltinUtil.h"
#include "Debug.h"

// *************************************************************************

#define TMP_REWRITE_FILE "/tmp/__rewrite.%s"
#define TMP_M_REWRITE_FILE "/tmp/__rewrite.%s.M"

// *************************************************************************
#ifdef    SAVE_INTER
static int unlink(char *path){

    static cnt = 0;
    char cmmd[100];
    *cerr_strm << "Not unlinking " << path << " (moving to rewsave." << cnt << ")\n";
    sprintf(cmmd,"/bin/mv %s /tmp/rewsave.%d\n", path, cnt++);
    system(cmmd);
    return(0);
}
#endif    // SAVE_INTER
// *************************************************************************
/*
    Called from yyparse() at the end of a module.  For each exported
    predicate, all the rules are written out to a temporary file,
    magic is called on it, and the magic output file is placed on
    the scanner input stack.
*/

void Cor_perform_rewriting(Cor_ParserStruct *Cor_parserStruct)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ExportInfo *export_info;
    
    if (Cor_parserStruct->CurModule->ExportedPreds == NULL) {
        *(Cor_exEnv->err_strm) << 
                "CORAL: Error - No exported query form.\n";
        return;
    }

    // create output file
    CoralStream *outf;
    int i;
    char tmp_rewrite[MAX_PATH_LEN];
    char tmp_m_rewrite[MAX_PATH_LEN];

    sprintf(tmp_rewrite, TMP_REWRITE_FILE, coral_pid);
    sprintf(tmp_m_rewrite, TMP_M_REWRITE_FILE, coral_pid);


    if (!(outf = Cor_open_outfile(tmp_rewrite))) {
        *(Cor_exEnv->err_strm) << 
            "CORAL::error : unable to open temporary file\n";
        *(Cor_exEnv->err_strm) << "Aborting attempt to process module\n";
        return;
    }

    // Process each exported query form independently.
    for (export_info = Cor_parserStruct->CurModule->ExportedPreds, i = 0;
        export_info != NULL; export_info = export_info->next, i++)
    {
        if (! Cor_parserStruct->CurModule->name.is_empty()) {
            *outf << "\nmodule "
                << Cor_parserStruct->CurModule->name.symbol() << i
                << ".\n";
        }
        *outf << "export " <<  export_info->pred_name.symbol() << "(";
        if (! export_info->adorn_symbol.is_empty())
            *outf << export_info->adorn_symbol.symbol();
        *outf << ").\n";

        Cor_parserStruct->CurModule->ConvertFunctions = 0;
        Cor_print_Annotations(*(Cor_parserStruct->CurModule), outf,
            Cor_parserStruct->make_index_annotations,
            Cor_parserStruct->agg_sel_annotations,
            Cor_parserStruct->AllowedAdornList,
            Cor_parserStruct->MultisetAnnotations,
            Cor_parserStruct->diskrel_annotations,
            Cor_parserStruct->asyncrel_annotations);

        Cor_Clause *rule;
        int j, num_rules;
        num_rules = Cor_parserStruct->cur_collect->count();

        for (j=0; j < num_rules; j++) {
            rule = Cor_parserStruct->cur_collect->get_ith(j);
            *outf << *rule;
        }

        // Add a print of the empty rule for each ASYNC_SOCK annotation

        for (Cor_PredAnnotations *anno= Cor_parserStruct->asyncrel_annotations;
                    anno; anno=anno->next() ) {
            anno->print(outf);
            *outf << " :- .\n";
        }

        *outf << "end_module.\n\n";
    }

    // close output file
    delete outf;

    char magic_input_file[200];
    char term_str_M[200];

    strcpy(magic_input_file, tmp_rewrite);

    // Now handle magic rewriting
    if (Cor_parseStack->cur_file_name())
        sprintf(term_str_M, "%s.M", Cor_parseStack->cur_file_name());
    else 
        sprintf(term_str_M, "%s.M", "stdin");

    if (Cor_exEnv->C_create_magic_default) {
        // check to see if the magic output file for the currently
        // consulted file already exists, else create it.
        if (!(outf = Cor_open_infile(term_str_M))) { 
            if (!(outf = Cor_open_outfile(term_str_M))) { 
                *(Cor_exEnv->err_strm) << 
                    "CORAL::error : unable to create temporary file\n";
                *(Cor_exEnv->err_strm) << 
                    "Aborting attempt to process module\n";
                unlink(magic_input_file);
                return ;
            }
            else
                delete outf;
        }
        else
            delete outf;
     }
    
    // perform magic rewriting on the file
    if (!Cor_callMagic(magic_input_file, tmp_rewrite)) {
        *(Cor_exEnv->err_strm) << "CORAL::error : unable to call magic\n";
        *(Cor_exEnv->err_strm) << "Aborting attempt to process module\n";
	*(Cor_exEnv->err_strm) << flush;
#ifdef    DEBUG
        *(Cor_exEnv->err_strm) << "Magic input file was: "
                               << magic_input_file << endl;
       // ASSERT(0);
#endif
        unlink(magic_input_file);
        return;
    }

    unlink(magic_input_file); 

    // open temporary magic output file and put the descriptor
    // on the scanner input stack
    if (!Cor_consultFile(tmp_m_rewrite, tmp_m_rewrite)) {
        CORAL_error(COR_FILE_NOT_FOUND, tmp_m_rewrite,
            "perform_rewriting", __FILE__, __LINE__);
        *(Cor_exEnv->err_strm) << "Aborting attempt to process module\n";
    }

    if (Cor_exEnv->C_create_magic_default) {
        // concatenate tmp_m_rewrite to the magic output file
        char buf[300];
        sprintf(buf, "cat %s >> %s", tmp_m_rewrite, term_str_M);
        if (system(buf)) {
             *(Cor_exEnv->err_strm) << 
                 "Unable to save magic output in " << term_str_M << endl;
        }
     }

    // delete tmp_m_rewrite
    unlink(tmp_m_rewrite);
}

/*****************************************************************************/

