// syntax: url_rel(local_name, arity, url, relation_name)

#include "Globals.h"
#include "CoralError.h"
#include "BuiltinRelation.h"
#include "TupleIterator.h"
#include "UrlRelation.h"

#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

Cor_BindEnv *UrlRelSolver(Cor_BuiltinRelation&, Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef URL_REL

    char *local_name, *url, *relation_name;
    int arity;
    
    if (iterator.arg_list.count() != 4) {
        CORAL_error(COR_BAD_ARITY,
            "4 arguments to url_rel needed", "UrlRelSolver");
        FAIL_RETURN;
    }
    
    Cor_Term term1 (iterator.arg_list[0], iterator.bindenv);
    Cor_deep_dereference_term (term1);
    Cor_Term term2 (iterator.arg_list[1], iterator.bindenv);
    Cor_deep_dereference_term (term2);
    Cor_Term term3 (iterator.arg_list[2], iterator.bindenv);
    Cor_deep_dereference_term (term3);
    Cor_Term term4 (iterator.arg_list[3], iterator.bindenv);
    Cor_deep_dereference_term (term4);

    if (!term1.expr.is_symbol()) {
        CORAL_error(COR_BAD_ARGLIST,
            "Local name must be a symbol","UrlRelSolver");
        term1.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        FAIL_RETURN;
    }
    if (!term2.expr.is_num()) {
        CORAL_error(COR_BAD_ARGLIST, "Arity must be a number","UrlRelSolver");
        term2.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        FAIL_RETURN;
    }
    if (!term3.expr.is_symbol()) {
        CORAL_error(COR_BAD_ARGLIST, "Url must be a symbol","UrlRelSolver");
        term3.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        FAIL_RETURN;
    }
    if (!term4.expr.is_symbol()) {
        CORAL_error(COR_BAD_ARGLIST,
            "Relation name must be a symbol","UrlRelSolver");
        term4.printon(Cor_exEnv->err_strm);
        *(Cor_exEnv->err_strm) << endl;
        FAIL_RETURN;
    }
    
    local_name = term1.expr.symbol();
    arity = term2.expr.NumIval();
    url = term3.expr.symbol();
    relation_name = term4.expr.symbol();
    
    Cor_Relation *new_rel;
    new_rel = Cor_find_external_relation(Cor_CArg(local_name),arity);
    if (new_rel) 
        *(Cor_exEnv->err_strm)
            << "Error: there is already local relation named "
            << local_name << endl;
    else {
        Cor_UrlRelation *url_rel = new Cor_UrlRelation(local_name,
            arity, url, relation_name);
        
        if (url_rel->Valid()) {
            Cor_CurDB->RelationTable->insert(Cor_CArg(local_name), url_rel);
            *(Cor_exEnv->out_strm)
                << "url relation " << local_name << " created.\n";
        }
        else
            *(Cor_exEnv->out_strm)
                << "url relation " << local_name << " cannot be created.\n";
    }
#endif // URL_REL

    return iterator.bindenv;
}
